"""Basic benchmark configuration."""

import logging
import sys
import time
from abc import ABC
from collections.abc import Callable
from datetime import datetime, timedelta, timezone
from typing import Generic, TypeVar

from attrs import define, field
from attrs.validators import instance_of
from cattrs import override
from cattrs.gen import make_dict_unstructure_fn
from pandas import DataFrame

from baybe.utils.random import temporary_seed
from benchmarks.result import Result, ResultMetadata
from benchmarks.serialization import BenchmarkSerialization, converter

logger = logging.getLogger(__name__)

stdout = logging.StreamHandler(stream=sys.stdout)

logger.addHandler(stdout)
logger.setLevel(logging.INFO)


@define(frozen=True, kw_only=True)
class BenchmarkSettings(ABC, BenchmarkSerialization):
    """The basic benchmark configuration."""

    random_seed: int = field(validator=instance_of(int), default=1337)
    """The used random seed."""


BenchmarkSettingsType = TypeVar("BenchmarkSettingsType", bound=BenchmarkSettings)


@define(frozen=True)
class Benchmark(Generic[BenchmarkSettingsType], BenchmarkSerialization):
    """The base class for all benchmark definitions."""

    function: Callable[[BenchmarkSettingsType], DataFrame] = field()
    """The callable containing the benchmarking logic."""

    settings: BenchmarkSettingsType = field()
    """The benchmark configuration."""

    @function.validator
    def _validate_function(self, _, function) -> None:
        if function.__doc__ is None:
            raise ValueError("The benchmark function must have a docstring.")

    @property
    def name(self) -> str:
        """The name of the benchmark function."""
        return self.function.__name__

    @property
    def description(self) -> str:
        """The description of the benchmark function."""
        assert self.function.__doc__ is not None
        return self.function.__doc__

    def __call__(self) -> Result:
        """Execute the benchmark and return the result."""
        start_datetime = datetime.now(timezone.utc)

        logger.info(
            "=" * 80
            + f"\nRunning benchmark '{self.name}' with "
            + f"random seed {self.settings.random_seed}.\n"
        )
        with temporary_seed(self.settings.random_seed):
            start_sec = time.perf_counter()
            result = self.function(self.settings)
            stop_sec = time.perf_counter()

        duration = timedelta(seconds=stop_sec - start_sec)

        logger.info(
            f"\nFinished benchmark '{self.name}' after {duration} "
            + f"with random seed {self.settings.random_seed}.\n"
            + "=" * 80
        )

        metadata = ResultMetadata(
            start_datetime=start_datetime,
            duration=duration,
        )

        return Result(self.name, result, metadata)


@converter.register_unstructure_hook
def unstructure_benchmark(benchmark: Benchmark) -> dict:
    """Unstructure a benchmark instance."""
    fn = make_dict_unstructure_fn(
        type(benchmark), converter, function=override(omit=True)
    )
    return {
        "name": benchmark.name,
        "description": benchmark.description,
        **fn(benchmark),
    }



## Bernoulli Multi Armed Bandit

# The goal of this example is twofold:
# - It shows how to use the {class}`~baybe.surrogates.bandit.BetaBernoulliMultiArmedBanditSurrogate` for use cases like A/B testing.
# - It demonstrates the difference between **optimization** and **active learning** in this context, by employing different acquisition functions chosen for each purpose.
#   Results are compared in terms of the achieved optimality of the resulting policy and the induced system identification error.

### Imports

import os
from dataclasses import dataclass
from typing import NamedTuple

import numpy as np
from scipy.stats import bernoulli

from baybe import Campaign
from baybe.acquisition import PosteriorStandardDeviation, qThompsonSampling
from baybe.acquisition.base import AcquisitionFunction
from baybe.parameters import CategoricalParameter
from baybe.recommenders import (
    BotorchRecommender,
    RandomRecommender,
    TwoPhaseMetaRecommender,
)
from baybe.surrogates import BetaBernoulliMultiArmedBanditSurrogate
from baybe.targets import BinaryTarget
from baybe.utils.random import set_random_seed

### An Imaginary Use Case

# Bandit models are a great choice when being faced with multiple (finite) options to choose from.
# Imagine yourself running a marketing campaign where you need to decide which version of marketing material to present to your customers.
# As in all other Bayesian optimization contexts, you are confronted with the classic exploration-exploitation dilemma:
# - **Exploitation:** On the one hand, you want to choose the most effective option right from the start.
# - **Exploration:** On the other hand, you cannot know in advance how effective each option is until you have tested it.
#
# Bandit models offer a principled approach to handling this dilemma.
# They can help you determine the effect size (e.g., the click-through rate) for your marketing options or optimize the effect size
# while the campaign is still running. Both goals are achieved by dynamically selecting the version of the marketing campaign that is shown to each customer.

### The Black Box Model

# Abstractly, we can think of our problem as a black box system with a finite number of actions that can be triggered.
# The image that comes to mind is that of a slot machine with multiple arms that can be pulled (hence the name "multi-armed bandit"), each offering a different win rate.
# You can think of each arm as representing a version of the marketing campaign, and the corresponding win rates would be the true effect size of each marketing material version.

# We can build such a black box system with just a few lines of code:


@dataclass
class MultiArmedBandit:
    """A stateless black box system offering a finite number of actions (= arms)."""

    win_rates: list[float]
    """Win rates associated with the arms."""

    def draw_arm(self, arm_index: int) -> bool:
        """Draw a reward by pulling the specified arm."""
        return bernoulli(self.win_rates[arm_index]).rvs()


### Setup

# Setting a random seed for reproducibility.

set_random_seed(42)

# For our example, we use a system with a fixed collection of win rates, which we
# assume are unknown to us:

WIN_RATES = [0.3, 0.4, 0.6, 0.65, 0.8]
BANDIT = MultiArmedBandit(WIN_RATES)

# Given this target system, we consider the following two tasks:
# - **Online optimization:** Maximizing the total reward accumulated during a campaign
# - **Active learning:** Estimating the win rates of all arms

# Solving these task can be achieved through an appropriate choice of acquisition function:
# For online optimization, we use {class}`~baybe.acquisition.acqfs.qThompsonSampling`.
# For active learning, we switch to {class}`~baybe.acquisition.acqfs.PosteriorStandardDeviation`.

# To estimate the corresponding effects, we simulate each campaign for a certain number of steps and repeat this process in multiple Monte Carlo runs:

SMOKE_TEST = "SMOKE_TEST" in os.environ

ACQFS = [
    qThompsonSampling(),  # Online optimization
    PosteriorStandardDeviation(),  # Active learning
]
N_MC_RUNS = 3 if SMOKE_TEST else 10
N_ITERATIONS = 50 if SMOKE_TEST else 200


### Building the Model

# To model the rewards obtained from the bandit, we use a {class}`~baybe.targets.binary.BinaryTarget`.
# The corresponding search space is spanned by a {class}`~baybe.parameters.categorical.CategoricalParameter` describing the bandit's arms.

n_arms = len(WIN_RATES)
target = BinaryTarget(name="clicked")
parameter = CategoricalParameter(
    name="Bandit Arm",
    values=[str(i) for i in range(n_arms)],
    # For bandits, one-hot parameter encoding is required:
    encoding="OHE",
)
surrogate = BetaBernoulliMultiArmedBanditSurrogate()


# For each simulation, we report the trajectory of earned rewards and the estimated win rates of the bandit arms:


class SimulationResult(NamedTuple):
    """Result of a bandit campaign simulation."""

    earned_rewards: np.ndarray
    estimated_win_rates: np.ndarray


# Putting all pieces together, we define a function to simulate our campaign under a given acquisition strategy:


def simulate(acqf: AcquisitionFunction) -> SimulationResult:
    """Simulate the campaign with with the given acquisition function."""
    recommender = TwoPhaseMetaRecommender(
        initial_recommender=RandomRecommender(),
        recommender=BotorchRecommender(
            surrogate_model=surrogate, acquisition_function=acqf
        ),
    )

    # Containers for storing the rewards and win rate estimates
    earned_rewards = np.zeros((N_MC_RUNS, N_ITERATIONS))
    estimated_win_rates = np.zeros((N_MC_RUNS, n_arms))

    for mc in range(N_MC_RUNS):
        searchspace = parameter.to_searchspace()
        objective = target.to_objective()
        campaign = Campaign(
            searchspace,
            objective,
            recommender,
            # The same arm can be pulled several times:
            allow_recommending_already_recommended=True,
            allow_recommending_already_measured=True,
        )

        for i in range(N_ITERATIONS):
            df = campaign.recommend(batch_size=1)
            recommended_arm_idx = int(df.values.item())
            reward = BANDIT.draw_arm(recommended_arm_idx)
            earned_rewards[mc, i] = reward
            df["clicked"] = reward
            campaign.add_measurements(df)

        estimated_win_rates[mc] = surrogate.posterior_means()

    return SimulationResult(earned_rewards, estimated_win_rates)


### Running the Simulation

# With our model in place, we can now simulate the campaign for both acquisition functions.
# The success of our two objectives is quantified using the following metrics:
# - For online optimization, we look at the **empirically achieved win rate** and
#   compare it to the win rate of the best arm.
# - For active learning, we report the **mean squared error of the win rates**
#   estimated at the end of the campaign.

for acqf in ACQFS:
    name = acqf.__class__.__name__
    result = simulate(acqf)
    estimates = result.estimated_win_rates
    estimation_bias = estimates.mean(axis=0) - WIN_RATES
    estimation_variance = estimates.var(axis=0)
    MSE = np.sum(estimation_bias**2 + estimation_variance)

    print(name)
    print("-" * len(name))
    print("Optimal win rate:", max(WIN_RATES))
    print("Achieved win rate:", result.earned_rewards.mean())
    print("Mean squared error (last iteration):", MSE)
    print("\n")


### Discussion

# As expected, we observe a higher achieved win rate for the online optimization strategy, as its underlying acquisition function is designed to recommend the arm with the largest reward most often.
# This comes at the cost of a higher estimation error for the remaining arms, which are sampled less frequently.
# In contrast, the active learning strategy attempts to balance the estimation errors across arms, resulting in a significantly smaller overall error.
# However, in order to do so, it must sample arms with low and high win rates equally often, which reduces the achieved rate and thus the total reward accumulated in the campaign.



"""SHAP insights."""

from __future__ import annotations

import warnings
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from attr.validators import deep_iterable
from attrs import define, field
from attrs.validators import instance_of
from shap import KernelExplainer

from baybe import Campaign
from baybe._optional.insights import shap
from baybe.exceptions import IncompatibleExplainerError, NoMeasurementsError
from baybe.objectives.base import Objective
from baybe.recommenders.base import RecommenderProtocol
from baybe.searchspace import SearchSpace
from baybe.surrogates import CompositeSurrogate
from baybe.surrogates.base import Surrogate, SurrogateProtocol
from baybe.utils.basic import is_all_instance
from baybe.utils.dataframe import to_tensor

_DEFAULT_EXPLAINER_CLS = "KernelExplainer"
SHAP_EXPLAINERS = {
    "AdditiveExplainer",
    "ExactExplainer",
    "KernelExplainer",
    "PartitionExplainer",
    "PermutationExplainer",
}
"""Supported SHAP explainer types for :class:`baybe.insights.shap.SHAPInsight`"""

NON_SHAP_EXPLAINERS = {"LimeTabular", "Maple"}
"""Supported non-SHAP explainer types for :class:`baybe.insights.shap.SHAPInsight`"""

EXPLAINERS = SHAP_EXPLAINERS | NON_SHAP_EXPLAINERS
"""Supported explainer types for :class:`baybe.insights.shap.SHAPInsight`"""

SHAP_PLOTS = {"bar", "beeswarm", "force", "heatmap", "scatter", "waterfall"}
"""Supported plot types for :meth:`baybe.insights.shap.SHAPInsight.plot`"""


def _get_explainer_cls(name: str) -> type[shap.Explainer]:
    """Retrieve the explainer class reference by name."""
    if name in SHAP_EXPLAINERS:
        return getattr(shap.explainers, name)
    if name in NON_SHAP_EXPLAINERS:
        return getattr(shap.explainers.other, name)
    raise ValueError(f"Unknown SHAP explainer class '{name}'.")


def is_shap_explainer(explainer: shap.Explainer) -> bool:
    """Indicate if the given explainer is a SHAP explainer or not (e.g. MAPLE, LIME)."""
    return type(explainer).__name__ in SHAP_EXPLAINERS


def make_explainer_for_surrogate(
    surrogate: Surrogate,
    data: pd.DataFrame,
    explainer_cls: type[shap.Explainer] | str = _DEFAULT_EXPLAINER_CLS,
    *,
    use_comp_rep: bool = False,
) -> shap.Explainer:
    """Create a SHAP explainer for a given surrogate model.

    Args:
        surrogate: The surrogate model to be explained.
        data: The background data set.
        explainer_cls: The SHAP explainer class for generating the explanation.
        use_comp_rep: Boolean flag specifying whether to explain the model's
            experimental or computational representation.

    Returns:
        The created explainer object.

    Raises:
        ValueError: If the provided background data set is empty.
        TypeError: If the provided explainer class is incompatible with the surrogate.
    """
    if data.empty:
        raise ValueError("The provided background data set is empty.")

    if isinstance(explainer_cls, str):
        explainer_cls = _get_explainer_cls(explainer_cls)

    if not (
        data.select_dtypes(exclude="number").empty
        or issubclass(explainer_cls, KernelExplainer)
    ):
        raise IncompatibleExplainerError(
            f"The selected explainer class '{explainer_cls.__name__}' does not support "
            f"categorical data. Switch to computational representation or use "
            f"'{KernelExplainer.__name__}'."
        )

    import torch

    if use_comp_rep:

        def model(x: np.ndarray) -> np.ndarray:
            tensor = to_tensor(x)
            with torch.no_grad():
                output = surrogate._posterior_comp(tensor).mean
            return output.numpy()

    else:

        def model(x: np.ndarray) -> np.ndarray:
            df = pd.DataFrame(x, columns=data.columns)
            with torch.no_grad():
                output = surrogate.posterior(df).mean
            return output.numpy()

    # Handle special settings: Lime default mode is otherwise set to "classification"
    kwargs = {"mode": "regression"} if explainer_cls.__name__ == "LimeTabular" else {}

    return explainer_cls(model, data, **kwargs)


@define
class SHAPInsight:
    """Class for SHAP-based feature importance insights.

    Also supports LIME and MAPLE explainers via the ``shap`` package.
    """

    explainers: tuple[shap.Explainer, ...] = field(
        converter=tuple,
        validator=deep_iterable(
            member_validator=instance_of(shap.Explainer),
        ),
    )
    """The explainer instances."""

    background_data: pd.DataFrame = field(validator=instance_of(pd.DataFrame))
    """The background data set used by the explainer."""

    @explainers.validator
    def _validate_explainers(self, _, explainers: tuple[shap.Explainer, ...]) -> None:
        """Validate the explainer type."""
        for explainer in explainers:
            if (name := explainer.__class__.__name__) not in EXPLAINERS:
                raise ValueError(
                    f"The given explainer type must be one of {EXPLAINERS}. "
                    f"Given: '{name}'."
                )

    @classmethod
    def from_surrogate(
        cls,
        surrogate: SurrogateProtocol,
        data: pd.DataFrame,
        explainer_cls: type[shap.Explainer] | str = _DEFAULT_EXPLAINER_CLS,
        *,
        use_comp_rep: bool = False,
    ) -> SHAPInsight:
        """Create a SHAP insight from a surrogate.

        For details, see :func:`make_explainer_for_surrogate`.
        """
        if isinstance(surrogate, Surrogate):
            single_output_surrogates = (surrogate,)
        elif isinstance(surrogate, CompositeSurrogate):
            single_output_surrogates = surrogate._surrogates_flat  # type:ignore[assignment]
            if not is_all_instance(single_output_surrogates, Surrogate):
                raise TypeError(
                    f"'{cls.__name__}.{cls.from_surrogate.__name__}' only supports "
                    f"'{CompositeSurrogate.__name__}' if it is composed only of models "
                    f"of type '{Surrogate.__name__}'."
                )
        else:
            raise ValueError(
                f"'{cls.__name__}.{cls.from_surrogate.__name__}' only accepts "
                f"surrogate models derived from '{Surrogate.__name__}' or "
                f"{CompositeSurrogate.__name__}."
            )

        explainers = tuple(
            make_explainer_for_surrogate(
                s, data, explainer_cls, use_comp_rep=use_comp_rep
            )
            for s in single_output_surrogates
        )
        return cls(explainers, data)

    @classmethod
    def from_campaign(
        cls,
        campaign: Campaign,
        explainer_cls: type[shap.Explainer] | str = _DEFAULT_EXPLAINER_CLS,
        *,
        use_comp_rep: bool = False,
    ) -> SHAPInsight:
        """Create a SHAP insight from a campaign.

        Uses the measurements of the campaign as background data.

        Args:
            campaign: A campaign using a surrogate-based recommender.
            explainer_cls: See :func:`make_explainer_for_surrogate`.
            use_comp_rep: See :func:`make_explainer_for_surrogate`.

        Returns:
            The SHAP insight object.

        Raises:
            ValueError: If the campaign does not contain any measurements.
        """
        if campaign.measurements.empty:
            raise NoMeasurementsError("The campaign does not contain any measurements.")
        data = campaign.measurements[[p.name for p in campaign.parameters]]
        background_data = campaign.searchspace.transform(data) if use_comp_rep else data

        return cls.from_surrogate(
            campaign.get_surrogate(),
            background_data,
            explainer_cls=explainer_cls,
            use_comp_rep=use_comp_rep,
        )

    @classmethod
    def from_recommender(
        cls,
        recommender: RecommenderProtocol,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        explainer_cls: type[shap.Explainer] | str = "KernelExplainer",
        *,
        use_comp_rep: bool = False,
    ) -> SHAPInsight:
        """Create a SHAP insight from a recommender.

        Uses the provided measurements to train the surrogate and as background data for
        the explainer.

        Args:
            recommender: A recommender using a surrogate model.
            searchspace: The searchspace for the recommender.
            objective: The objective for the recommender.
            measurements: The measurements for training the surrogate and the explainer.
            explainer_cls: See :func:`make_explainer_for_surrogate`.
            use_comp_rep: See :func:`make_explainer_for_surrogate`.

        Returns:
            The SHAP insight object.

        Raises:
            TypeError: If the recommender has no ``get_surrogate`` method.
        """
        if not hasattr(recommender, "get_surrogate"):
            raise TypeError(
                f"The provided recommender does not provide a surrogate model. "
                f"'{cls.__name__}' needs a surrogate model and thus only works with "
                f"model-based recommenders."
            )

        surrogate_model = recommender.get_surrogate(
            searchspace, objective, measurements
        )

        data = measurements[[p.name for p in searchspace.parameters]]

        return cls.from_surrogate(
            surrogate_model,
            searchspace.transform(data) if use_comp_rep else data,
            explainer_cls=explainer_cls,
            use_comp_rep=use_comp_rep,
        )

    def explain_target(
        self, target_index: int, data: pd.DataFrame | None = None, /
    ) -> shap.Explanation:
        """Compute Shapley explanations for a given data set for a single-target.

        Args:
            target_index: The index of the target for which the explanation is created.
            data: The dataframe for which the Shapley values are to be computed.
                By default, the background data set of the explainer is used.

        Returns:
            The computed Shapley explanation.

        Raises:
            ValueError: If not all the columns of the explainer background dataframe
                are present in the given data.
        """
        if data is None:
            data = self.background_data
        elif not set(self.background_data.columns).issubset(data.columns):
            raise ValueError(
                "The provided dataframe must contain all columns that were used for "
                "the background data."
            )

        # Align columns with background data
        df_aligned = data[self.background_data.columns]

        explainer = self.explainers[target_index]
        if not is_shap_explainer(explainer):
            # Return attributions for non-SHAP explainers
            if explainer.__module__.endswith("maple"):
                # Additional argument for maple to increase comparability to SHAP
                attributions = explainer.attributions(
                    df_aligned, multiply_by_input=True
                )[0]
            else:
                attributions = explainer.attributions(df_aligned)[0]

            explanation = shap.Explanation(
                values=attributions,
                base_values=explainer.model(self.background_data).mean(),
                data=df_aligned.values,
                feature_names=df_aligned.columns.values,
            )
        else:
            explanation = explainer(df_aligned)

        # Permute explanation object data according to input column order
        # (`base_values` can be a scalar or vector)
        # TODO: https://github.com/shap/shap/issues/3958
        idx = self.background_data.columns.get_indexer(data.columns)
        idx = idx[idx != -1]  # Additional columns in data are ignored.
        for attr in ["values", "data", "base_values"]:
            try:
                setattr(explanation, attr, getattr(explanation, attr)[:, idx])
            except IndexError as ex:
                if not (
                    isinstance(explanation.base_values, float)
                    or explanation.base_values.shape[1] == 1
                ):
                    raise TypeError("Unexpected explanation format.") from ex
        explanation.feature_names = [explanation.feature_names[i] for i in idx]

        # Reduce dimensionality of explanations to 2D in case
        # a 3D explanation is returned. This is the case for
        # some explainers even if only one output is present.
        if len(explanation.shape) == 3:
            if explanation.shape[2] == 1:
                # Some explainers have a third dimension corresponding to the
                # number of model outputs (in this implementation always 1).
                explanation = explanation[:, :, 0]
            else:
                # Some explainers (like ``AdditiveExplainer``) have a third
                # dimension corresponding to feature interactions. The total shap
                # value is obtained by summing over them.
                explanation = explanation.sum(axis=2)
        elif len(explanation.shape) != 2:
            raise RuntimeError(
                f"The explanation obtained for '{self.__class__.__name__}' has an "
                f"unexpected dimensionality of {len(explanation.shape)}."
            )

        return explanation

    def explain(
        self, data: pd.DataFrame | None = None, /
    ) -> tuple[shap.Explanation, ...]:
        """Compute Shapley explanations for a given data set for all targets.

        Args:
            data: The dataframe for which the Shapley values are to be computed.
                By default, the background data set of the explainer is used.

        Returns:
            The computed Shapley explanations.
        """
        return tuple(self.explain_target(k, data) for k in range(len(self.explainers)))

    def plot(
        self,
        plot_type: Literal[
            "bar", "beeswarm", "force", "heatmap", "scatter", "waterfall"
        ],
        data: pd.DataFrame | None = None,
        /,
        *,
        show: bool = True,
        explanation_index: int | None = None,
        target_index: int | None = None,
        **kwargs: Any,
    ) -> plt.Axes:
        """Plot the Shapley values using the provided plot type.

        Args:
            plot_type: The type of plot to be created.
            data: See :meth:`explain`.
            show: Boolean flag determining if the plot is to be rendered.
            explanation_index: Positional index of the data point that should be
                explained. Only relevant for plot types that can only handle a single
                data point.
            target_index: The index of the target for which the plot is created. Only
                required for multi-output objectives.
            **kwargs: Additional keyword arguments passed to the plot function.

        Returns:
            The plot object.

        Raises:
            ValueError: If the provided plot type is not supported.
            ValueError: If the target index is not specified for multi-output
                situations.
        """
        if data is None:
            data = self.background_data
        if target_index is None:
            if len(self.explainers) > 1:
                raise ValueError(
                    "The 'target_index' must be specified for multi-output scenarios."
                )
            target_index = 0

        # Use custom scatter plot function to ignore non-numeric features
        if plot_type == "scatter":
            return self._plot_shap_scatter(data, target_index, show=show, **kwargs)

        if plot_type not in SHAP_PLOTS:
            raise ValueError(
                f"Invalid plot type: '{plot_type}'. Available options: {SHAP_PLOTS}."
            )
        plot_func = getattr(shap.plots, plot_type)

        # Handle plot types that only explain a single data point
        if plot_type in ["force", "waterfall"]:
            if explanation_index is None:
                warnings.warn(
                    f"When using plot type '{plot_type}', an 'explanation_index' must "
                    f"be chosen to identify a single data point that should be "
                    f"explained. Choosing the first entry at position 0."
                )
                explanation_index = 0

            toplot = self.explain_target(target_index, data.iloc[[explanation_index]])
            toplot = toplot[0]

            if plot_type == "force":
                kwargs["matplotlib"] = True
        else:
            toplot = self.explain_target(target_index, data)

        return plot_func(toplot, show=show, **kwargs)

    def _plot_shap_scatter(
        self,
        data: pd.DataFrame | None = None,
        target_index: int = 0,
        /,
        *,
        show: bool = True,
        **kwargs: Any,
    ) -> plt.Axes:
        """Plot the Shapley values as scatter plot, ignoring non-numeric features.

        For details, see :meth:`explain`.
        """
        if data is None:
            data = self.background_data

        df_numeric = data.select_dtypes("number")
        numeric_idx = data.columns.get_indexer(df_numeric.columns)
        if df_numeric.empty:
            raise ValueError(
                "No SHAP scatter plot can be created since all features contain "
                "non-numeric values."
            )
        if non_numeric_cols := set(data.columns) - set(df_numeric.columns):
            warnings.warn(
                f"The following features are excluded from the SHAP scatter plot "
                f"because they contain non-numeric values: {non_numeric_cols}",
                UserWarning,
            )
        return shap.plots.scatter(
            self.explain_target(target_index, data)[:, numeric_idx], show=show, **kwargs
        )



"""Constraint utilities."""

import numpy as np
import pandas as pd

from baybe.parameters.utils import is_inactive
from baybe.searchspace import SubspaceContinuous


def is_cardinality_fulfilled(
    df: pd.DataFrame,
    subspace_continuous: SubspaceContinuous,
    *,
    check_minimum: bool = True,
    check_maximum: bool = True,
) -> bool:
    """Validate cardinality constraints in a dataframe of parameter configurations.

    Args:
        df: The dataframe to be checked.
        subspace_continuous: The subspace spanned by the considered parameters.
        check_minimum: If ``True``, minimum cardinality constraints are validated.
        check_maximum: If ``True``, maximum cardinality constraints are validated.

    Returns:
        ``True`` if all cardinality constraints are fulfilled, ``False`` otherwise.
    """
    for c in subspace_continuous.constraints_cardinality:
        # Get the activity thresholds for all parameters
        cols = df[c.parameters]
        thresholds = {
            p.name: c.get_absolute_thresholds(p.bounds)
            for p in subspace_continuous.get_parameters_by_name(c.parameters)
        }
        lower_thresholds = [thresholds[p].lower for p in cols.columns]
        upper_thresholds = [thresholds[p].upper for p in cols.columns]

        # Count the number of active values per dataframe row
        inactives = is_inactive(cols, lower_thresholds, upper_thresholds)
        n_zeros = inactives.sum(axis=1)
        n_active = len(c.parameters) - n_zeros

        # Check if cardinality is violated
        if check_minimum and np.any(n_active < c.min_cardinality):
            return False
        if check_maximum and np.any(n_active > c.max_cardinality):
            return False

    return True



"""Easom function with transfer learning, negated function and a noise_std of 0.05."""

from __future__ import annotations

import math

import numpy as np
import pandas as pd

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import NumericalDiscreteParameter, TaskParameter
from baybe.parameters.base import DiscreteParameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.definition import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)


def easom(x: np.ndarray, noise_std: float = 0.0, negate: bool = False):
    """Eason function output values.

    Args:
        x: Input with two features
        noise_std: Noise to be added to output.
        negate: Whether to invert the output

    Returns:
        Easom function output
    """
    x = np.array(x).ravel()
    assert x.shape == (2,)
    y = (
        -math.cos(x[0])
        * math.cos(x[1])
        * math.exp(-((x[0] - math.pi) ** 2) - (x[1] - math.pi) ** 2)
    )
    if negate:
        y = y * -1
    if noise_std > 0:
        y += np.random.normal(loc=0.0, scale=noise_std, size=1)[0]
    return y


def easom_tl_47_negate_noise5(settings: ConvergenceBenchmarkSettings) -> pd.DataFrame:
    """Benchmark function for transfer learning with the Easom function.

    Key characteristics:
    • Compares two negated versions of Easom function:
      - Target: standard negated Easom
      - Source: negated Easom with added noise (noise_std=0.05)
    • Uses 47 points per dimension
    • Tests transfer learning with different source data percentages:
      - 1% of source data
      - 5% of source data
      - 10% of source data
      - 20% of source data

    Args:
        settings: Configuration settings for the convergence benchmark

    Returns:
        DataFrame containing benchmark results
    """
    points_per_dim = 47
    negate = True
    percentages = [0.01, 0.05, 0.1]
    bounds: np.ndarray = np.array([[-10] * 2, [10] * 2])
    grid_locations = {
        f"x{d}": np.linspace(lower, upper, points_per_dim)
        for d, (lower, upper) in enumerate(bounds.T)
    }

    functions = {
        "Target_Function": lambda x: easom(x, negate=negate),
        "Source_Function": lambda x: easom(x, noise_std=0.05, negate=negate),
    }

    params: list[DiscreteParameter] = [
        NumericalDiscreteParameter(name=name, values=values)
        for name, values in grid_locations.items()
    ]
    task_param = TaskParameter(
        name="Function",
        values=["Target_Function", "Source_Function"],
        active_values=["Target_Function"],
    )
    params_tl = params + [task_param]

    searchspace_nontl = SearchSpace.from_product(parameters=params)
    searchspace_tl = SearchSpace.from_product(parameters=params_tl)

    objective = SingleTargetObjective(
        target=NumericalTarget(name="Target", mode="MAX" if negate else "MIN")
    )
    tl_campaign = Campaign(
        searchspace=searchspace_tl,
        objective=objective,
    )
    nontl_campaign = Campaign(
        searchspace=searchspace_nontl,
        objective=objective,
    )

    meshgrid = np.meshgrid(*[points for points in grid_locations.values()])

    lookups = []
    for function_name, function in functions.items():
        lookup = pd.DataFrame(
            {f"x{d}": grid_d.ravel() for d, grid_d in enumerate(meshgrid)}
        )
        lookup["Target"] = lookup.apply(function, axis=1)
        lookup["Function"] = function_name
        lookups.append(lookup)
    concat_lookups = pd.concat(lookups)

    lookup = concat_lookups[concat_lookups["Function"] == "Target_Function"]
    initial_data = concat_lookups[concat_lookups["Function"] == "Source_Function"]

    results = []
    for p in percentages:
        results.append(
            simulate_scenarios(
                {
                    f"{int(100 * p)}": tl_campaign,
                    f"{int(100 * p)}_naive": nontl_campaign,
                },
                lookup,
                initial_data=[
                    initial_data.sample(frac=p) for _ in range(settings.n_mc_iterations)
                ],
                batch_size=settings.batch_size,
                n_doe_iterations=settings.n_doe_iterations,
                impute_mode="error",
            )
        )
    results.append(
        simulate_scenarios(
            {"0": tl_campaign, "0_naive": nontl_campaign},
            lookup,
            batch_size=settings.batch_size,
            n_doe_iterations=settings.n_doe_iterations,
            n_mc_iterations=settings.n_mc_iterations,
            impute_mode="error",
        )
    )
    return pd.concat(results)


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=2,
    n_doe_iterations=30,
    n_mc_iterations=100,
)

easom_tl_47_negate_noise5_benchmark = ConvergenceBenchmark(
    function=easom_tl_47_negate_noise5,
    optimal_target_values={"Target": 0.9635009628660742},
    settings=benchmark_config,
)



## Modeling a Mixture in Slot-Based Representation

### Terminology

# Modeling a mixture is possible in a non-traditional way by using a concept we refer to
# as a **slot**. A slot is represented through the combination of two parameters: one
# indicating the *amount* of a mixture ingredient, and another indicating the *type* of
# the ingredient (as a label) populating the slot. Unlike in [traditional mixture
# modeling](/examples/Mixtures/traditional.md), the total number of parameters is not
# determined by how many ingredient choices we have, but by the maximum number of slots
# we allow. For instance, if we want to design a mixture with *up to three* ingredients,
# we can do so by creating three slots represented by six parameters.

# A corresponding search space could look like this:
# | Slot1_Label | Slot1_Amount | Slot2_Label | Slot2_Amount | Slot3_Label | Slot3_Amount |
# |:------------|:-------------|:------------|:-------------|:------------|:-------------|
# | Solvent1    | 10           | Solvent5    | 20           | Solvent4    | 70           |
# | Solvent1    | 30           | Solvent8    | 40           | Solvent2    | 30           |
# | Solvent3    | 20           | Solvent1    | 35           | Solvent9    | 45           |
# | Solvent2    | 15           | Solvent3    | 40           | Solvent1    | 45           |

# The slot-based representation has one decided advantage over traditional
# modeling: We can use BayBE's label encodings for the label parameters. For
# instance, when mixing small molecules, the
# [`SubstanceParameter`](baybe.parameters.substance.SubstanceParameter) can be used to
# smartly encode the slot labels, enabling the algorithm to perform a chemically-aware
# mixture optimization.

# In this example, we show how to design such a search space, including the various
# discrete constraints we need to impose. We consider a situation where we want to mix
# up to three solvents, whose respective amounts must add up to 100.

# ```{admonition} Discrete vs. Continuous Modeling
# :class: important
#
# Here, we only use discrete parameters, although in principle the parameters
# corresponding to amounts could also be modeled as continuous numbers. However, this
# would imply that some of the constraints would have to act on both discrete and
# continuous parameters, which is not currently supported.
# ```

### Imports

import math

import numpy as np
import pandas as pd

from baybe.constraints import (
    DiscreteDependenciesConstraint,
    DiscreteNoLabelDuplicatesConstraint,
    DiscretePermutationInvarianceConstraint,
    DiscreteSumConstraint,
    ThresholdCondition,
)
from baybe.parameters import NumericalDiscreteParameter, SubstanceParameter
from baybe.searchspace import SubspaceDiscrete
from baybe.utils.dataframe import pretty_print_df

# Basic example settings:

SUM_TOLERANCE = 0.1  # tolerance allowed to fulfill the sum constraints
RESOLUTION = 5  # resolution for discretizing the slot amounts

### Parameter Setup

# First, we create the parameters for the slot labels. Each slot offers a choice of
# four solvents:

dict_solvents = {
    "water": "O",
    "ethanol": "CCO",
    "methanol": "CO",
    "acetone": "CC(=O)C",
}
slot1_label = SubstanceParameter(
    name="Slot1_Label", data=dict_solvents, encoding="MORDRED"
)
slot2_label = SubstanceParameter(
    name="Slot2_Label", data=dict_solvents, encoding="MORDRED"
)
slot3_label = SubstanceParameter(
    name="Slot3_Label", data=dict_solvents, encoding="MORDRED"
)

# Next, we create the parameters representing the slot amounts:

slot1_amount = NumericalDiscreteParameter(
    name="Slot1_Amount", values=np.linspace(0, 100, RESOLUTION), tolerance=0.2
)
slot2_amount = NumericalDiscreteParameter(
    name="Slot2_Amount", values=np.linspace(0, 100, RESOLUTION), tolerance=0.2
)
slot3_amount = NumericalDiscreteParameter(
    name="Slot3_Amount", values=np.linspace(0, 100, RESOLUTION), tolerance=0.2
)

# We collect all parameters in a single list:

parameters = [
    slot1_label,
    slot2_label,
    slot3_label,
    slot1_amount,
    slot2_amount,
    slot3_amount,
]

### Constraint Setup

# For the sake of demonstration, we consider a scenario where we do *not* care about the
# order of addition of components to the mixture, which imposes two additional
# constraints: one for removing duplicates and one for imposing permutation invariance.
#
# ```{admonition} Order of Addition
# :class: note
# Whether you need to impose the constraints for removing duplicates and imposing
# permutation invariance depends on your use case. If the order of addition is relevant
# to your mixture, the permutation invariance constraint should be discarded and one
# could further argue that adding the same substance multiple times should be allowed.
# ```

#### Duplicate Substances

# Assuming that the order of addition is irrelevant, there is no difference between
# having two slots with the same substance or having only one slot with the combined
# amounts. Thus, we want to make sure that there are no such duplicate label entries,
# which can be achieved using a
# {class}`~baybe.constraints.discrete.DiscreteNoLabelDuplicatesConstraint`:

no_duplicates_constraint = DiscreteNoLabelDuplicatesConstraint(
    parameters=["Slot1_Label", "Slot2_Label", "Slot3_Label"]
)

#### Permutation Invariance

# Next, we need to take care of permutation invariance. If our order of addition does
# not matter, the result of interchanging any two slots does not alter the overall
# mixture, i.e. the mixture slots are considered permutation-invariant.

# A complication with permutation invariance arises from the fact that we do not only
# have a label per slot, but also a numerical amount. If this amount is zero, then the
# label of the slot becomes meaningless, because adding zero of the corresponding
# substance does not change the mixture. In BayBE, we call this a "dependency", i.e.
# the slot labels depend on the slot amounts and are only relevant if the amount
# satisfies some condition (in this case "amount > 0").

# The {class}`~baybe.constraints.discrete.DiscreteDependenciesConstraint` informs the
# {class}`~baybe.constraints.discrete.DiscretePermutationInvarianceConstraint` about
# these dependencies so that they are correctly included in the filtering process:

perm_inv_constraint = DiscretePermutationInvarianceConstraint(
    parameters=["Slot1_Label", "Slot2_Label", "Slot3_Label"],
    dependencies=DiscreteDependenciesConstraint(
        parameters=["Slot1_Amount", "Slot2_Amount", "Slot3_Amount"],
        conditions=[
            ThresholdCondition(threshold=0.0, operator=">"),
            ThresholdCondition(threshold=0.0, operator=">"),
            ThresholdCondition(threshold=0.0, operator=">"),
        ],
        affected_parameters=[["Slot1_Label"], ["Slot2_Label"], ["Slot3_Label"]],
    ),
)

#### Substance Amounts

# Interpreting the slot amounts as percentages, we need to ensure that their total is
# always 100:

sum_constraint = DiscreteSumConstraint(
    parameters=["Slot1_Amount", "Slot2_Amount", "Slot3_Amount"],
    condition=ThresholdCondition(threshold=100, operator="=", tolerance=SUM_TOLERANCE),
)


# We store all constraints in a single list:

constraints = [perm_inv_constraint, sum_constraint, no_duplicates_constraint]


### Search Space Creation

# With all building blocks in place, we can now assemble our discrete space and inspect
# its configurations:


space = SubspaceDiscrete.from_product(parameters=parameters, constraints=constraints)
print(
    pretty_print_df(
        space.exp_rep,
        max_rows=len(space.exp_rep),
        max_columns=len(space.exp_rep.columns),
    )
)

# ````{admonition} Simplex Construction
# :class: tip
# In this example, we use the
# {meth}`~baybe.searchspace.discrete.SubspaceDiscrete.from_product` constructor in order
# to demonstrate the explicit creation of all involved constraints. However, for
# creating mixture representations, the
# {meth}`~baybe.searchspace.discrete.SubspaceDiscrete.from_simplex` constructor should
# generally be used. It takes care of the overall sum constraint already during search
# space creation, providing a more efficient path to the same result.
#
# The alternative in our case would look like:
# ```python
# space = SubspaceDiscrete.from_simplex(
#     max_sum=100.0,
#     boundary_only=True,
#     simplex_parameters=[slot1_amount, slot2_amount, slot3_amount],
#     product_parameters=[slot1_label, slot2_label, slot3_label],
#     constraints=[perm_inv_constraint, no_duplicates_constraint],
# )
# ```
# Note that {meth}`~baybe.searchspace.discrete.SubspaceDiscrete.from_simplex`
# inherently ensures the sum constraint, hence we do not pass it to `constraints`.
# ````


### Verification of Constraints

# Let us programmatically assert that all constraints are satisfied:

amounts = space.exp_rep[["Slot1_Amount", "Slot2_Amount", "Slot3_Amount"]]
labels = space.exp_rep[["Slot1_Label", "Slot2_Label", "Slot3_Label"]]
slots = space.exp_rep.apply(
    lambda row: pd.Series(
        [(row[f"Slot{k}_Label"], row[f"Slot{k}_Amount"]) for k in range(1, 4)]
    ),
    axis=1,
)

# * All amounts sum to 100:

n_wrong_sum = amounts.sum(axis=1).apply(lambda x: x - 100).abs().gt(SUM_TOLERANCE).sum()
assert n_wrong_sum == 0
print("Number of configurations whose amounts do not sum to 100: ", n_wrong_sum)


# * There are no duplicate slot labels:

n_duplicates = labels.nunique(axis=1).ne(3).sum()
assert n_duplicates == 0
print("Number of configurations with duplicate slot labels: ", n_duplicates)


# * There are no permutation-invariant configurations:

n_permute = slots.apply(frozenset, axis=1).duplicated().sum()
assert n_permute == 0
print("Number of permuted configurations: ", n_permute)

### Verification of Span

# Finally, we also assert if we have completely spanned the space of allowed
# configurations by comparing the numbers of unique `K`-solvent entries against their
# theoretical values.

# ```{admonition} Theoretical Span
# :class: info
#
# The number of possible `K`-solvent entries can be found by imagining the corresponding
# [traditional mixture representation](/examples/Mixtures/traditional.md) and solving a
# slightly more complex version of the ["stars and bars"
# problem](https://en.wikipedia.org/wiki/Stars_and_bars_(combinatorics)), where the
# number of non-empty bins is fixed. That is, we need to ask how many possible ways
# exist to distribute `N` items (= number of elemental steps for the amounts, in our
# case `RESOLUTION-1`) across `M` bins (= number of available solvents) if exactly
# `K` bins are non-empty (= number of solvents allowed in the mixture).
#
# There are `(M choose K)` ways to select the non-empty buckets. When distributing the
# `N` items, one item needs to go to each of the `K` buckets for it to be non-empty.
# The remaining `N - K` items can be freely distributed among the `K` buckets. The
# number of configurations for the latter is given by the "stars and bars" formula,
# which states that `X` indistinguishable items can be placed in `Y` distinguishable
# bins in `((X + Y -1) choose (Y - 1))` ways. Setting `X`=`N-K` and `Y`=`K` gives
# `((N - 1) choose (K - 1))`. Combined with the former count, we get the formula
# implemented in the helper function below.
# ```

# Helper function to compute the theoretical numbers:


def n_combinations(N: int, M: int, K: int) -> int:
    """Get number of ways to put `N` items into `M` bins yielding `K` non-empty bins."""
    return math.comb(M, K) * math.comb(N - 1, K - 1)


# Verify that the space is fully spanned:

for K in range(1, 4):
    n_combinations_expected = n_combinations(RESOLUTION - 1, len(dict_solvents), K)
    n_combinations_actual = (amounts != 0).sum(axis=1).eq(K).sum()
    assert n_combinations_expected == n_combinations_actual
    print(
        f"Number of unique {K}-solvent entries: "
        f"{n_combinations_actual} ({n_combinations_expected} expected)"
    )



"""Synthetic function with two continuous and one discrete input."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
import pandas as pd
from numpy import pi, sin, sqrt
from pandas import DataFrame

from baybe.campaign import Campaign
from baybe.parameters import NumericalContinuousParameter, NumericalDiscreteParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.definition.convergence import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)

if TYPE_CHECKING:
    from mpl_toolkits.mplot3d import Axes3D


def _lookup(arr: np.ndarray, /) -> np.ndarray:
    """Numpy-based lookup callable defining the objective function."""
    x, y, z = np.array_split(arr, 3, axis=1)
    try:
        assert np.all(-2 * pi <= x) and np.all(x <= 2 * pi)
        assert np.all(-2 * pi <= y) and np.all(y <= 2 * pi)
        assert np.all(np.isin(z, [1, 2, 3, 4]))
    except AssertionError:
        raise ValueError("Inputs are not in the valid ranges.")

    return (
        (z == 1) * sin(x) * (1 + sin(y))
        + (z == 2) * (x * sin(0.9 * x) + sin(x) * sin(y))
        + (z == 3) * (sqrt(x + 8) * sin(x) + sin(x) * sin(y))
        + (z == 4) * (x * sin(1.666 * sqrt(x + 8)) + sin(x) * sin(y))
    )


def lookup(df: pd.DataFrame, /) -> pd.DataFrame:
    """Dataframe-based lookup callable used as the loop-closing element."""
    return pd.DataFrame(
        _lookup(df[["x", "y", "z"]].to_numpy()), columns=["target"], index=df.index
    )


def synthetic_2C1D_1C(settings: ConvergenceBenchmarkSettings) -> DataFrame:
    """Benchmark function with hybrid synthetic test function.

    Key characteristics:
    • Parameters:
      - z: Discrete {1,2,3,4}
      - x: Continuous [-2π, 2π]
      - y: Continuous [-2π, 2π]
    • Output: Continuous
    • Objective: Maximization
    • Optimal inputs:
      - {x: 1.610, y: 1.571, z: 3}
      - {x: 1.610, y: -4.712, z: 3}
    • Optimal output: 4.09685
    • Tests multiple recommenders:
      - Random Recommender
      - Default Recommender

    Args:
        settings: Configuration settings for the convergence benchmark.

    Returns:
        DataFrame containing benchmark results.
    """
    parameters = [
        NumericalContinuousParameter("x", (-2 * pi, 2 * pi)),
        NumericalContinuousParameter("y", (-2 * pi, 2 * pi)),
        NumericalDiscreteParameter("z", (1, 2, 3, 4)),
    ]

    target = NumericalTarget(name="target", mode="MAX")
    searchspace = SearchSpace.from_product(parameters=parameters)
    objective = target.to_objective()

    scenarios: dict[str, Campaign] = {
        "Random Recommender": Campaign(
            searchspace=searchspace,
            recommender=RandomRecommender(),
            objective=objective,
        ),
        "Default Recommender": Campaign(
            searchspace=searchspace,
            objective=objective,
        ),
    }

    return simulate_scenarios(
        scenarios,
        lookup,
        batch_size=settings.batch_size,
        n_doe_iterations=settings.n_doe_iterations,
        n_mc_iterations=settings.n_mc_iterations,
        impute_mode="error",
    )


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=5,
    n_doe_iterations=40,
    n_mc_iterations=100,
)

synthetic_2C1D_1C_benchmark = ConvergenceBenchmark(
    function=synthetic_2C1D_1C,
    optimal_target_values={"target": 4.09685},
    settings=benchmark_config,
)


if __name__ == "__main__":
    #  Visualize the domain

    import matplotlib.pyplot as plt

    X = np.linspace(-2 * pi, 2 * pi)
    Y = np.linspace(-2 * pi, 2 * pi)
    Z = [1, 2, 3, 4]

    x_mesh, y_mesh = np.meshgrid(X, Y)

    fig = plt.figure(figsize=(10, 10))
    for i, z in enumerate(Z):
        ax: Axes3D = fig.add_subplot(2, 2, i + 1, projection="3d")
        t_mesh = _lookup(
            np.c_[x_mesh.ravel(), y_mesh.ravel(), np.repeat(z, x_mesh.size)]
        ).reshape(x_mesh.shape)
        ax.plot_surface(x_mesh, y_mesh, t_mesh)
        plt.title(f"{z=}")

    plt.show()



"""Test alternative ways of creation not considered in the strategies."""

import pytest

from baybe.acquisition.base import AcquisitionFunction
from baybe.recommenders import BotorchRecommender
from baybe.utils.basic import get_subclasses

abbreviation_list = [
    cl.abbreviation
    for cl in get_subclasses(AcquisitionFunction)
    if hasattr(cl, "abbreviation")
]

fullname_list = [cl.__name__ for cl in get_subclasses(AcquisitionFunction)]


@pytest.mark.parametrize("acqf", abbreviation_list + fullname_list)
def test_creation_from_string(acqf):
    """Tests the creation from strings."""
    AcquisitionFunction.from_dict({"type": acqf})


@pytest.mark.parametrize("acqf", abbreviation_list + fullname_list)
def test_string_usage_in_recommender(acqf):
    """Tests the recommender initialization with acqfs as string."""
    BotorchRecommender(acquisition_function=acqf)



"""A collection of point sampling algorithms."""

import warnings
from enum import Enum
from typing import Literal

import numpy as np
import pandas as pd


def farthest_point_sampling(
    points: np.ndarray,
    n_samples: int = 1,
    initialization: Literal["farthest", "random"] = "farthest",
    random_tie_break: bool = True,
) -> list[int]:
    """Select a subset of points using farthest point sampling.

    Creates a subset of a given collection of points by successively adding points with
    the largest Euclidean distance to intermediate point selections encountered during
    the algorithmic process. The mechanism used for the initial point selection is
    configurable.

    Args:
        points: The points that are available for selection, represented as a 2-D array
            of shape ``(n, k)``, where ``n`` is the number of points and ``k`` is the
            dimensionality of the points.
        n_samples: The total number of points to be selected.
        initialization: Determines how the first points are selected:
            * ``"farthest"``: The first two selected points are those with the
              largest distance. If only a single point is requested, a deterministic
              choice is made based on the point coordinates.
            * ``"random"``: The first point is selected uniformly at random.
        random_tie_break: Determines if points are chosen deterministically or randomly
            in equidistant situations. If ``True``, a random point is selected from the
            candidates, otherwise the first point is selected. For non-equidistant
            points, the point with the largest minimum distance is always selected.

    Returns:
        A list containing the positional indices of the selected points.

    Raises:
        ValueError: If the provided array is not two-dimensional.
        ValueError: If the array contains no points.
        ValueError: If the input space has no dimensions.
        ValueError: If an unknown method for initialization is specified.
    """
    if (n_dims := np.ndim(points)) != 2:
        raise ValueError(
            f"The provided array must be two-dimensional but the given input had "
            f"{n_dims} dimensions."
        )
    if (n_points := len(points)) == 0:
        raise ValueError("The provided array must contain at least one row.")
    if points.shape[-1] == 0:
        raise ValueError("The provided input space must be at least one-dimensional.")
    if n_samples > n_points:
        raise ValueError(
            f"The number of requested samples ({n_samples}) cannot be larger than the "
            f"total number of points provided ({n_points})."
        )

    # Catch the pathological case upfront
    if len(np.unique(points, axis=0)) == 1:
        warnings.warn("All points are identical.", UserWarning)
        return list(range(n_samples))

    # Sort the points to produce the same result regardless of the input order
    sort_idx = np.lexsort(tuple(points.T))
    points = points[sort_idx]

    # Pre-compute the pairwise distances between all points
    from sklearn.metrics import pairwise_distances

    dist_matrix = pairwise_distances(points)

    # Avoid wrong behavior situations where all (remaining) points are duplicates
    np.fill_diagonal(dist_matrix, -np.inf)

    # Initialize the point selection
    if initialization == "random":
        selected_point_indices = [np.random.randint(0, n_points)]
    elif initialization == "farthest":
        idx_1d = np.argmax(dist_matrix)
        selected_point_indices = list(
            map(int, np.unravel_index(idx_1d, dist_matrix.shape))
        )
        if n_samples == 1:
            return [sort_idx[selected_point_indices[0]]]
    else:
        raise ValueError(f"unknown initialization recommender: '{initialization}'")

    # Initialize the list of remaining points
    remaining_point_indices = list(range(n_points))
    for idx in selected_point_indices:
        remaining_point_indices.remove(idx)

    # Successively add the points with the largest distance
    while len(selected_point_indices) < n_samples:
        # Collect distances between selected and remaining points
        dist = dist_matrix[np.ix_(remaining_point_indices, selected_point_indices)]

        # Find for each candidate point the smallest distance to the selected points
        min_dists = np.min(dist, axis=1)

        if random_tie_break:
            # Select a random point that has the "largest smallest distance"
            max_val = np.max(min_dists)
            max_indices = np.where(min_dists == max_val)[0]
            choice = np.random.choice(max_indices)
        else:
            # Choose the first point with the "largest smallest distance"
            choice = np.argmax(min_dists)
        selected_point_index = remaining_point_indices[choice]

        # Add the chosen point to the selection
        selected_point_indices.append(selected_point_index)
        remaining_point_indices.remove(selected_point_index)

    # Undo the initial point reordering
    return sort_idx[selected_point_indices].tolist()


class DiscreteSamplingMethod(Enum):
    """Available discrete sampling methods."""

    Random = "Random"
    """Random Sampling."""

    FPS = "FPS"
    """Farthest point sampling."""


def sample_numerical_df(
    df: pd.DataFrame,
    n_points: int,
    *,
    method: DiscreteSamplingMethod = DiscreteSamplingMethod.Random,
) -> pd.DataFrame:
    """Sample data points from a numerical dataframe.

    If the requested amount of points is larger than the number of available points,
    the entire dataframe will be returned as many times at it fits into the requested
    number and the specified sampling method will only return the remainder of points.

    Args:
        df: Dataframe with purely numerical entries.
        n_points: Number of points to sample.
        method: Sampling method.

    Returns:
        The sampled points.

    Raises:
        TypeError: If the provided dataframe has non-numerical content.
        ValueError: When an invalid sampling method was provided.
    """
    if any(df[col].dtype.kind not in "iufb" for col in df.columns):
        raise TypeError(
            f"'{sample_numerical_df.__name__}' only supports purely numerical "
            f"dataframes."
        )

    # Split points in trivial and sampled parts
    n_trivial, n_sampled = divmod(n_points, len(df))

    ilocs = list(range(len(df))) * n_trivial
    if n_sampled > 0:
        if method is DiscreteSamplingMethod.FPS:
            ilocs += farthest_point_sampling(df.values, n_sampled)
        elif method is DiscreteSamplingMethod.Random:
            ilocs += df.reset_index(drop=True).sample(n_sampled).index.tolist()
        else:
            raise ValueError(f"Unrecognized sampling method: '{method}'.")

    return df.iloc[ilocs]



"""Collection of basic kernels."""

import gc

from attrs import define, field
from attrs.converters import optional as optional_c
from attrs.validators import ge, gt, in_, instance_of
from attrs.validators import optional as optional_v
from typing_extensions import override

from baybe.kernels.base import BasicKernel
from baybe.priors.base import Prior
from baybe.utils.conversion import fraction_to_float
from baybe.utils.validation import finite_float


@define(frozen=True)
class LinearKernel(BasicKernel):
    """A linear kernel."""

    variance_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel variance parameter."""

    variance_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel variance parameter."""

    @override
    def to_gpytorch(self, *args, **kwargs):
        import torch

        from baybe.utils.torch import DTypeFloatTorch

        gpytorch_kernel = super().to_gpytorch(*args, **kwargs)
        if (initial_value := self.variance_initial_value) is not None:
            gpytorch_kernel.variance = torch.tensor(
                initial_value, dtype=DTypeFloatTorch
            )
        return gpytorch_kernel


@define(frozen=True)
class MaternKernel(BasicKernel):
    """A Matern kernel using a smoothness parameter."""

    nu: float = field(
        converter=fraction_to_float, validator=in_([0.5, 1.5, 2.5]), default=2.5
    )
    """A smoothness parameter.

    Only takes the values 0.5, 1.5 or 2.5. Larger values yield smoother interpolations.
    """

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""


@define(frozen=True)
class PeriodicKernel(BasicKernel):
    """A periodic kernel."""

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""

    period_length_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel period length."""

    period_length_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel period length."""

    @override
    def to_gpytorch(self, *args, **kwargs):
        import torch

        from baybe.utils.torch import DTypeFloatTorch

        gpytorch_kernel = super().to_gpytorch(*args, **kwargs)
        # lengthscale is handled by the base class

        if (initial_value := self.period_length_initial_value) is not None:
            gpytorch_kernel.period_length = torch.tensor(
                initial_value, dtype=DTypeFloatTorch
            )
        return gpytorch_kernel


@define(frozen=True)
class PiecewisePolynomialKernel(BasicKernel):
    """A piecewise polynomial kernel."""

    q: int = field(validator=in_([0, 1, 2, 3]), default=2)
    """A smoothness parameter."""

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""


@define(frozen=True)
class PolynomialKernel(BasicKernel):
    """A polynomial kernel."""

    power: int = field(validator=[instance_of(int), ge(0)])
    """The power of the polynomial term."""

    offset_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel offset."""

    offset_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel offset."""

    @override
    def to_gpytorch(self, *args, **kwargs):
        import torch

        from baybe.utils.torch import DTypeFloatTorch

        gpytorch_kernel = super().to_gpytorch(*args, **kwargs)
        if (initial_value := self.offset_initial_value) is not None:
            gpytorch_kernel.offset = torch.tensor(initial_value, dtype=DTypeFloatTorch)
        return gpytorch_kernel


@define(frozen=True)
class RBFKernel(BasicKernel):
    """A radial basis function (RBF) kernel."""

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""


@define(frozen=True)
class RFFKernel(BasicKernel):
    """A random Fourier features (RFF) kernel."""

    num_samples: int = field(validator=[instance_of(int), ge(1)])
    """The number of frequencies to draw."""

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""


@define(frozen=True)
class RQKernel(BasicKernel):
    """A rational quadratic (RQ) kernel."""

    lengthscale_prior: Prior | None = field(
        default=None, validator=optional_v(instance_of(Prior))
    )
    """An optional prior on the kernel lengthscale."""

    lengthscale_initial_value: float | None = field(
        default=None,
        converter=optional_c(float),
        validator=optional_v([finite_float, gt(0.0)]),
    )
    """An optional initial value for the kernel lengthscale."""


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Validation functionality for surrogates."""

from __future__ import annotations

from collections.abc import Callable
from typing import Any

import cattrs
from cattrs import ClassValidationError
from cattrs.strategies import configure_union_passthrough

from baybe.surrogates.base import Surrogate


def validate_custom_architecture_cls(model_cls: type) -> None:
    """Validate a custom architecture to have the correct attributes.

    Args:
        model_cls: The user defined model class.

    Raises:
        ValueError: When model_cls does not have _fit or _posterior.
        ValueError: When _fit or _posterior is not a callable method.
        ValueError: When _fit does not have the required signature.
        ValueError: When _posterior does not have the required signature.
    """
    # Methods must exist
    if not (hasattr(model_cls, "_fit") and hasattr(model_cls, "_posterior")):
        raise ValueError(
            "`_fit` and a `_posterior` must exist for custom architectures"
        )

    fit = model_cls._fit
    posterior = model_cls._posterior

    # They must be methods
    if not (callable(fit) and callable(posterior)):
        raise ValueError(
            "`_fit` and a `_posterior` must be methods for custom architectures"
        )

    # Methods must have the correct arguments
    params = fit.__code__.co_varnames[: fit.__code__.co_argcount]

    if params != Surrogate._fit.__code__.co_varnames:
        raise ValueError(
            "Invalid args in `_fit` method definition for custom architecture. "
            "Please refer to Surrogate._fit for the required function signature."
        )

    params = posterior.__code__.co_varnames[: posterior.__code__.co_argcount]

    if params != Surrogate._posterior.__code__.co_varnames:
        raise ValueError(
            "Invalid args in `_posterior` method definition for custom architecture. "
            "Please refer to Surrogate._posterior for the required function signature."
        )


# Create a strict type validation converter
type_validation_converter = cattrs.Converter(forbid_extra_keys=True)
"""Converter used for strict type validation."""

configure_union_passthrough(int | float | str | None, type_validation_converter)


@type_validation_converter.register_structure_hook
def _strict_int_structure_hook(obj: Any, _: type[int]) -> int:
    if isinstance(obj, int) and not isinstance(obj, bool):  # Exclude bools
        return obj
    raise ValueError(
        f"Value '{obj}' (type: {type(obj).__name__}) is not a valid integer. "
        "Only actual 'int' instances are accepted."
    )


@type_validation_converter.register_structure_hook
def _strict_float_structure_hook(obj: Any, _: type[float]) -> float:
    if isinstance(obj, float):
        return obj
    raise ValueError(
        f"Value '{obj}' (type: {type(obj).__name__}) is not a valid float. "
        "Only actual 'float' instances are accepted."
    )


@type_validation_converter.register_structure_hook
def _strict_bool_structure_hook(obj: Any, _: type[bool]) -> bool:
    if isinstance(obj, bool):
        return obj
    raise ValueError(
        f"Value '{obj}' (type: {type(obj).__name__}) is not a valid boolean. "
        "Only actual 'bool' instances (True, False) are accepted."
    )


def make_dict_validator(specification: type) -> Callable:
    """Construct an attrs dictionary validator based on a ``TypedDict``.

    Args:
        specification: Describes allowed keys and corresponding value types.

    Returns:
        An attrs compatible validator.
    """

    def validate_model_params(_instance: Any, attr: Any, value: dict) -> None:
        """Validate attrs attribute using cattrs with an extremely strict int hook."""
        try:
            type_validation_converter.structure(value, specification)
        except ClassValidationError as ex:
            raise TypeError(
                f"The provided dictionary for '{attr.name}' is invalid."
            ) from ex

    return validate_model_params



"""Functionality to "simulate" Bayesian DOE given a lookup mechanism.

The term "simulation" can have two slightly different interpretations, depending on the
applied context:

*   It can refer to "backtesting" a particular DOE strategy on a fixed (finite)
    dataset. In this context, "simulation" means investigating what experimental
    trajectory we would have observed had we applied the recommender in a certain
    defined context and restricted the possible parameter configurations to those
    contained in the dataset.

*   It can refer to the simulation of an actual DOE loop (i.e. recommending experiments
    and retrieving the corresponding measurements) where the loop closure is realized
    in the form of a callable (black-box) function that can be queried during the
    optimization to provide target values.
"""

from baybe.simulation.core import simulate_experiment
from baybe.simulation.scenarios import simulate_scenarios
from baybe.simulation.transfer_learning import simulate_transfer_learning

__all__ = [
    "simulate_experiment",
    "simulate_scenarios",
    "simulate_transfer_learning",
]



"""Functionality for managing DOE campaigns. Main point of interaction via Python."""

from __future__ import annotations

import gc
import json
from collections.abc import Callable, Collection, Sequence
from functools import reduce
from typing import TYPE_CHECKING, Any

import cattrs
import numpy as np
import pandas as pd
from attrs import Attribute, Factory, define, evolve, field, fields
from attrs.converters import optional
from attrs.validators import instance_of
from typing_extensions import override

from baybe.constraints.base import DiscreteConstraint
from baybe.exceptions import (
    IncompatibilityError,
    NoMeasurementsError,
    NotEnoughPointsLeftError,
)
from baybe.objectives.base import Objective, to_objective
from baybe.parameters.base import Parameter
from baybe.recommenders.base import RecommenderProtocol
from baybe.recommenders.meta.base import MetaRecommender
from baybe.recommenders.meta.sequential import TwoPhaseMetaRecommender
from baybe.recommenders.pure.bayesian.base import BayesianRecommender
from baybe.recommenders.pure.nonpredictive.base import NonPredictiveRecommender
from baybe.searchspace._filtered import FilteredSubspaceDiscrete
from baybe.searchspace.core import (
    SearchSpace,
    SearchSpaceType,
    to_searchspace,
    validate_searchspace_from_config,
)
from baybe.serialization import SerialMixin, converter
from baybe.surrogates.base import PosteriorStatistic, SurrogateProtocol
from baybe.targets.base import Target
from baybe.telemetry import (
    TELEM_LABELS,
    telemetry_record_recommended_measurement_percentage,
    telemetry_record_value,
)
from baybe.utils.basic import UNSPECIFIED, UnspecifiedType, is_all_instance
from baybe.utils.boolean import eq_dataframe
from baybe.utils.conversion import to_string
from baybe.utils.dataframe import (
    _ValidatedDataFrame,
    filter_df,
    fuzzy_row_match,
    normalize_input_dtypes,
)
from baybe.utils.validation import validate_parameter_input, validate_target_input

if TYPE_CHECKING:
    from botorch.acquisition import AcquisitionFunction
    from botorch.posteriors import Posterior

# Metadata columns
_RECOMMENDED = "recommended"
_MEASURED = "measured"
_EXCLUDED = "excluded"
_METADATA_COLUMNS = [_RECOMMENDED, _MEASURED, _EXCLUDED]


def _make_allow_flag_default_factory(
    default: bool,
) -> Callable[[Campaign], bool | UnspecifiedType]:
    """Make a default factory for allow_* flags."""

    def default_allow_flag(campaign: Campaign) -> bool | UnspecifiedType:
        """Attrs-compatible default factory for allow_* flags."""
        if campaign.searchspace.type is SearchSpaceType.DISCRETE:
            return default
        return UNSPECIFIED

    return default_allow_flag


def _validate_allow_flag(campaign: Campaign, attribute: Attribute, value: Any) -> None:
    """Attrs-compatible validator for context-aware validation of allow_* flags."""
    match campaign.searchspace.type:
        case SearchSpaceType.DISCRETE:
            if not isinstance(value, bool):
                raise ValueError(
                    f"For search spaces of '{SearchSpaceType.DISCRETE}', "
                    f"'{attribute.name}' must be a Boolean."
                )
        case _:
            if value is not UNSPECIFIED:
                raise ValueError(
                    f"For search spaces of type other than "
                    f"'{SearchSpaceType.DISCRETE}', '{attribute.name}' cannot be set "
                    f"since the flag is meaningless in such contexts.",
                )


@define
class Campaign(SerialMixin):
    """Main class for interaction with BayBE.

    Campaigns define and record an experimentation process, i.e. the execution of a
    series of measurements and the iterative sequence of events involved.

    In particular, a campaign:
        * Defines the objective of an experimentation process.
        * Defines the search space over which the experimental parameter may vary.
        * Defines a recommender for exploring the search space.
        * Records the measurement data collected during the process.
        * Records metadata about the progress of the experimentation process.
    """

    # DOE specifications
    searchspace: SearchSpace = field(converter=to_searchspace)
    """The search space in which the experiments are conducted.
    When passing a :class:`baybe.parameters.base.Parameter`,
    a :class:`baybe.searchspace.discrete.SubspaceDiscrete`, or a
    a :class:`baybe.searchspace.continuous.SubspaceContinuous`, conversion to
    :class:`baybe.searchspace.core.SearchSpace` is automatically applied."""

    objective: Objective | None = field(default=None, converter=optional(to_objective))
    """The optimization objective.
    When passing a :class:`baybe.targets.base.Target`, conversion to
    :class:`baybe.objectives.single.SingleTargetObjective` is automatically applied."""

    recommender: RecommenderProtocol = field(
        factory=TwoPhaseMetaRecommender,
        validator=instance_of(RecommenderProtocol),
    )
    """The employed recommender"""

    allow_recommending_already_measured: bool | UnspecifiedType = field(
        default=Factory(
            _make_allow_flag_default_factory(default=True), takes_self=True
        ),
        validator=_validate_allow_flag,
        kw_only=True,
    )
    """Allow to recommend experiments that were already measured earlier.
    Can only be set for discrete search spaces."""

    allow_recommending_already_recommended: bool | UnspecifiedType = field(
        default=Factory(
            _make_allow_flag_default_factory(default=False), takes_self=True
        ),
        validator=_validate_allow_flag,
        kw_only=True,
    )
    """Allow to recommend experiments that were already recommended earlier.
    Can only be set for discrete search spaces."""

    allow_recommending_pending_experiments: bool | UnspecifiedType = field(
        default=Factory(
            _make_allow_flag_default_factory(default=False), takes_self=True
        ),
        validator=_validate_allow_flag,
        kw_only=True,
    )
    """Allow pending experiments to be part of the recommendations.
    Can only be set for discrete search spaces."""

    # Metadata
    _searchspace_metadata: pd.DataFrame = field(init=False, eq=eq_dataframe)
    """Metadata tracking the experimentation status of the search space."""

    n_batches_done: int = field(default=0, init=False)
    """The number of already processed batches."""

    n_fits_done: int = field(default=0, init=False)
    """The number of fits already done."""

    # Private
    _measurements_exp: pd.DataFrame = field(
        factory=pd.DataFrame, eq=eq_dataframe, init=False
    )
    """The experimental representation of the conducted experiments."""

    _cached_recommendation: pd.DataFrame = field(
        factory=pd.DataFrame, eq=eq_dataframe, init=False
    )
    """The cached recommendations."""

    @_searchspace_metadata.default
    def _default_searchspace_metadata(self) -> pd.DataFrame:
        """Create a fresh metadata object."""
        df = pd.DataFrame(
            False,
            index=self.searchspace.discrete.exp_rep.index,
            columns=_METADATA_COLUMNS,
        )

        return df

    @override
    def __str__(self) -> str:
        recommended_count = sum(self._searchspace_metadata[_RECOMMENDED])
        measured_count = sum(self._searchspace_metadata[_MEASURED])
        excluded_count = sum(self._searchspace_metadata[_EXCLUDED])
        n_elements = len(self._searchspace_metadata)
        searchspace_fields = [
            to_string(
                "Recommended:",
                f"{recommended_count}/{n_elements}",
                single_line=True,
            ),
            to_string(
                "Measured:",
                f"{measured_count}/{n_elements}",
                single_line=True,
            ),
            to_string(
                "Excluded:",
                f"{excluded_count}/{n_elements}",
                single_line=True,
            ),
        ]
        metadata_fields = [
            to_string("Batches done", self.n_batches_done, single_line=True),
            to_string("Fits done", self.n_fits_done, single_line=True),
            to_string("Discrete Subspace Meta Data", *searchspace_fields),
        ]
        metadata = to_string("Meta Data", *metadata_fields)
        fields = [metadata, self.searchspace, self.objective, self.recommender]

        return to_string(self.__class__.__name__, *fields)

    @property
    def measurements(self) -> pd.DataFrame:
        """The experimental data added to the Campaign."""
        return self._measurements_exp

    @property
    def parameters(self) -> tuple[Parameter, ...]:
        """The parameters of the underlying search space."""
        return self.searchspace.parameters

    @property
    def targets(self) -> tuple[Target, ...]:
        """The targets of the underlying objective."""
        return self.objective.targets if self.objective is not None else ()

    @classmethod
    def from_config(cls, config_json: str) -> Campaign:
        """Create a campaign from a configuration JSON.

        Args:
            config_json: The string with the configuration JSON.

        Returns:
            The constructed campaign.
        """
        config = json.loads(config_json)
        return converter.structure(config, Campaign)

    @classmethod
    def validate_config(cls, config_json: str) -> None:
        """Validate a given campaign configuration JSON.

        Args:
            config_json: The JSON that should be validated.
        """
        config = json.loads(config_json)
        _validation_converter.structure(config, Campaign)

    def add_measurements(
        self,
        data: pd.DataFrame,
        numerical_measurements_must_be_within_tolerance: bool = True,
    ) -> None:
        """Add results from a dataframe to the internal database.

        Each addition of data is considered a new batch. Added results are checked for
        validity. Categorical values need to have an exact match. For numerical values,
        a campaign flag determines if values that lie outside a specified tolerance
        are accepted. Possible validation exceptions are documented in
        :func:`baybe.utils.validation.validate_target_input` and
        :func:`baybe.utils.validation.validate_parameter_input`.

        Args:
            data: The data to be added (with filled values for targets). Preferably
                created via :func:`baybe.campaign.Campaign.recommend`.
            numerical_measurements_must_be_within_tolerance: Flag indicating if
                numerical parameters need to be within their tolerances.
        """
        # Invalidate recommendation cache first (in case of uncaught exceptions below)
        self._cached_recommendation = pd.DataFrame()

        # Validate target and parameter input values
        validate_target_input(data, self.targets)
        validate_parameter_input(
            data, self.parameters, numerical_measurements_must_be_within_tolerance
        )
        data = normalize_input_dtypes(data, self.parameters + self.targets)
        data.__class__ = _ValidatedDataFrame

        # Read in measurements and add them to the database
        self.n_batches_done += 1
        to_insert = data.copy()
        to_insert["BatchNr"] = self.n_batches_done
        to_insert["FitNr"] = np.nan

        self._measurements_exp = pd.concat(
            [self._measurements_exp, to_insert], axis=0, ignore_index=True
        )

        # Update metadata
        if self.searchspace.type in (SearchSpaceType.DISCRETE, SearchSpaceType.HYBRID):
            idxs_matched = fuzzy_row_match(
                self.searchspace.discrete.exp_rep, data, self.parameters
            )
            self._searchspace_metadata.loc[idxs_matched, _MEASURED] = True

        # Telemetry
        telemetry_record_value(TELEM_LABELS["COUNT_ADD_RESULTS"], 1)
        telemetry_record_recommended_measurement_percentage(
            self._cached_recommendation, data, self.parameters
        )

    def update_measurements(
        self,
        data: pd.DataFrame,
        numerical_measurements_must_be_within_tolerance: bool = True,
    ) -> None:
        """Update previously added measurements.

        This can be useful to correct mistakes or update target measurements. The
        match to existing data entries is made based on the index. This will reset
        the `FitNr` of the corresponding measurement and reset cached recommendations.

        Args:
            data: The measurement data to be updated (with filled values for targets).
            numerical_measurements_must_be_within_tolerance: Flag indicating if
                numerical parameters need to be within their tolerances.

        Raises:
            ValueError: If the given data contains duplicated indices.
            ValueError: If the given data contains indices not present in existing
                measurements.
        """
        # Validate target and parameter input values
        validate_target_input(data, self.targets)
        validate_parameter_input(
            data, self.parameters, numerical_measurements_must_be_within_tolerance
        )
        data = normalize_input_dtypes(data, self.parameters + self.targets)
        data.__class__ = _ValidatedDataFrame

        # Block duplicate input indices
        if data.index.has_duplicates:
            raise ValueError(
                "The input dataframe containing the measurement updates has duplicated "
                "indices. Please ensure that all updates for a given measurement are "
                "made in a single combined entry."
            )

        # Allow only existing indices
        if nonmatching_idxs := set(data.index).difference(self._measurements_exp.index):
            raise ValueError(
                f"Updating measurements requires indices matching the "
                f"existing measurements. The following indices were in the input, but "
                f"are not found in the existing entries: {nonmatching_idxs}"
            )

        # Perform the update
        cols = [p.name for p in self.parameters] + [t.name for t in self.targets]
        self._measurements_exp.loc[data.index, cols] = data[cols]

        # Reset fit number and cached recommendations
        self._measurements_exp.loc[data.index, "FitNr"] = np.nan
        self._cached_recommendation = pd.DataFrame()

    def toggle_discrete_candidates(  # noqa: DOC501
        self,
        constraints: Collection[DiscreteConstraint] | pd.DataFrame,
        exclude: bool,
        complement: bool = False,
        dry_run: bool = False,
    ) -> pd.DataFrame:
        """In-/exclude certain discrete points in/from the candidate set.

        Args:
            constraints: A filtering mechanism determining the candidates subset to be
                in-/excluded. Can be either a collection of
                :class:`~baybe.constraints.base.DiscreteConstraint` or a dataframe.
                For the latter, see :func:`~baybe.utils.dataframe.filter_df`
                for details.
            exclude: If ``True``, the specified candidates are excluded.
                If ``False``, the candidates are considered for recommendation.
            complement: If ``True``, the filtering mechanism is inverted so that
                the complement of the candidate subset specified by the filter is
                toggled. For details, see :func:`~baybe.utils.dataframe.filter_df`.
            dry_run: If ``True``, the target subset is only extracted but not
                affected. If ``False``, the candidate set is updated correspondingly.
                Useful for setting up the correct filtering mechanism.

        Returns:
            A new dataframe containing the  discrete candidate set passing through the
            specified filter.
        """
        # Clear cache
        self._cached_recommendation = pd.DataFrame()

        df = self.searchspace.discrete.exp_rep

        if isinstance(constraints, pd.DataFrame):
            # Determine the candidate subset to be toggled
            points = filter_df(df, constraints, complement)

        elif isinstance(constraints, Collection) and is_all_instance(
            constraints, DiscreteConstraint
        ):
            # TODO: Should be taken over by upcoming `SubspaceDiscrete.filter` method,
            #   automatically choosing the appropriate backend (polars/pandas/...)

            # Filter the search space dataframe according to the given constraint
            idx = reduce(
                lambda x, y: x.intersection(y), (c.get_valid(df) for c in constraints)
            )

            # Determine the candidate subset to be toggled
            points = df.drop(index=idx) if complement else df.loc[idx].copy()

        else:
            raise TypeError(
                "Candidate toggling is not implemented for the given type of "
                "constraint specifications."
            )

        if not dry_run:
            self._searchspace_metadata.loc[points.index, _EXCLUDED] = exclude

        return points

    def recommend(
        self,
        batch_size: int,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        """Provide the recommendations for the next batch of experiments.

        Args:
            batch_size: Number of requested recommendations.
            pending_experiments: Parameter configurations specifying experiments
                that are currently pending.

        Returns:
            Dataframe containing the recommendations in experimental representation.

        Raises:
            ValueError: If ``batch_size`` is smaller than 1.
        """
        if batch_size < 1:
            raise ValueError(
                f"You must at least request one recommendation per batch, but provided "
                f"{batch_size=}."
            )

        # Invalidate cached recommendation if pending experiments are provided
        if (pending_experiments is not None) and not pending_experiments.empty:
            self._cached_recommendation = pd.DataFrame()
            validate_parameter_input(pending_experiments, self.parameters)
            pending_experiments = normalize_input_dtypes(
                pending_experiments, self.parameters
            )
            pending_experiments.__class__ = _ValidatedDataFrame

        # If there are cached recommendations and the batch size of those is equal to
        # the previously requested one, we just return those
        if len(self._cached_recommendation) == batch_size:
            return self._cached_recommendation

        # Update recommendation meta data
        if len(self._measurements_exp) > 0:
            self.n_fits_done += 1
            self._measurements_exp.fillna({"FitNr": self.n_fits_done}, inplace=True)

        # Prepare the search space according to the current campaign state
        if self.searchspace.type is SearchSpaceType.DISCRETE:
            # TODO: This implementation should at some point be hidden behind an
            #   appropriate public interface, like `SubspaceDiscrete.filter()`
            mask_todrop = self._searchspace_metadata[_EXCLUDED].copy()
            if not self.allow_recommending_already_recommended:
                mask_todrop |= self._searchspace_metadata[_RECOMMENDED]
            if not self.allow_recommending_already_measured:
                mask_todrop |= self._searchspace_metadata[_MEASURED]
            if (
                not self.allow_recommending_pending_experiments
                and pending_experiments is not None
            ):
                mask_todrop |= pd.merge(
                    self.searchspace.discrete.exp_rep,
                    pending_experiments,
                    indicator=True,
                    how="left",
                )["_merge"].eq("both")
            searchspace = evolve(
                self.searchspace,
                discrete=FilteredSubspaceDiscrete.from_subspace(
                    self.searchspace.discrete, ~mask_todrop.to_numpy()
                ),
            )
        else:
            searchspace = self.searchspace

        # Pending experiments should not be passed to non-predictive recommenders
        # to avoid complaints about unused arguments, so we need to know of what
        # type the next recommender will be
        recommender = self.recommender
        if isinstance(recommender, MetaRecommender):
            recommender = recommender.get_non_meta_recommender(
                batch_size,
                searchspace,
                self.objective,
                self._measurements_exp,
                pending_experiments,
            )
        is_nonpredictive = isinstance(recommender, NonPredictiveRecommender)

        # Get the recommended search space entries
        try:
            # NOTE: The `recommend` call must happen on `self.recommender` to update
            #   potential inner states in case of meta recommenders!
            rec = self.recommender.recommend(
                batch_size,
                searchspace,
                self.objective,
                self._measurements_exp,
                None if is_nonpredictive else pending_experiments,
            )
        except NotEnoughPointsLeftError as ex:
            # Aliases for code compactness
            f = fields(Campaign)
            ok_m = self.allow_recommending_already_measured
            ok_r = self.allow_recommending_already_recommended
            ok_p = self.allow_recommending_pending_experiments
            ok_m_name = f.allow_recommending_already_measured.name
            ok_r_name = f.allow_recommending_already_recommended.name
            ok_p_name = f.allow_recommending_pending_experiments.name
            no_blocked_pending_points = ok_p or (pending_experiments is None)

            # If there are no candidate restrictions to be relaxed
            if ok_m and ok_r and no_blocked_pending_points:
                raise ex

            # Otherwise, extract possible relaxations
            solution = [
                f"'{name}=True'"
                for name, value in [
                    (ok_m_name, ok_m),
                    (ok_r_name, ok_r),
                    (ok_p_name, no_blocked_pending_points),
                ]
                if not value
            ]
            message = solution[0] if len(solution) == 1 else " and/or ".join(solution)
            raise NotEnoughPointsLeftError(
                f"{str(ex)} Consider setting {message}."
            ) from ex

        # Cache the recommendations
        self._cached_recommendation = rec.copy()

        # Update metadata
        if self.searchspace.type in (SearchSpaceType.DISCRETE, SearchSpaceType.HYBRID):
            self._searchspace_metadata.loc[rec.index, _RECOMMENDED] = True

        # Telemetry
        telemetry_record_value(TELEM_LABELS["COUNT_RECOMMEND"], 1)
        telemetry_record_value(TELEM_LABELS["BATCH_SIZE"], batch_size)

        return rec

    def posterior(self, candidates: pd.DataFrame | None = None) -> Posterior:
        """Get the posterior predictive distribution for the given candidates.

        Args:
            candidates: The candidate points in experimental recommendations. If not
                provided, the posterior for the existing campaign measurements is
                returned. For details, see
                :meth:`baybe.surrogates.base.Surrogate.posterior`.

        Raises:
            IncompatibilityError: If the underlying surrogate model exposes no
                method for computing the posterior distribution.

        Returns:
            Posterior: The corresponding posterior object.
            For details, see :meth:`baybe.surrogates.base.Surrogate.posterior`.
        """
        if candidates is None:
            candidates = self.measurements[[p.name for p in self.parameters]]

        surrogate = self.get_surrogate()
        if not hasattr(surrogate, method_name := "posterior"):
            raise IncompatibilityError(
                f"The used surrogate type '{surrogate.__class__.__name__}' does not "
                f"provide a '{method_name}' method."
            )

        return surrogate.posterior(candidates)

    def posterior_stats(
        self,
        candidates: pd.DataFrame | None = None,
        stats: Sequence[PosteriorStatistic] = ("mean", "std"),
    ) -> pd.DataFrame:
        """Return posterior statistics for each target.

        Args:
            candidates: The candidate points in experimental representation. If not
                provided, the statistics of the existing campaign measurements are
                calculated. For details, see
                :meth:`baybe.surrogates.base.Surrogate.posterior_stats`.
            stats: Sequence indicating which statistics to compute. Also accepts
                floats, for which the corresponding quantile point will be computed.

        Raises:
            ValueError: If a requested quantile is outside the open interval (0,1).
            TypeError: If the posterior utilized by the surrogate does not support
                a requested statistic.

        Returns:
            A dataframe with posterior statistics for each target and candidate.
        """
        if candidates is None:
            if self.measurements.empty:
                raise NoMeasurementsError(
                    f"No candidates were provided and the campaign has no measurements "
                    f"yet. '{self.posterior_stats.__name__}' has no candidates to "
                    f"compute statistics for in this case."
                )

            candidates = self.measurements[[p.name for p in self.parameters]]

        surrogate = self.get_surrogate()
        if not hasattr(surrogate, method_name := "posterior_stats"):
            raise IncompatibilityError(
                f"The used surrogate type '{surrogate.__class__.__name__}' does not "
                f"provide a '{method_name}' method."
            )

        return surrogate.posterior_stats(candidates, stats)

    def get_surrogate(
        self,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> SurrogateProtocol:
        """Get the current surrogate model.

        Args:
            batch_size: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.
            pending_experiments: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.

        Raises:
            IncompatibilityError: If the current recommender does not provide a
                surrogate model.

        Returns:
            Surrogate: The surrogate of the current recommender.

        Note:
            Currently, this method always returns the surrogate model with respect to
            the transformed target(s) / objective. This means that if you are using a
            ``SingleTargetObjective`` with a transformed target or a
            ``DesirabilityObjective``, the model's output will correspond to the
            transformed quantities and not the original untransformed target(s).
        """
        if self.objective is None:
            raise IncompatibilityError(
                f"No surrogate is available since no '{Objective.__name__}' is defined."
            )

        recommender = self._get_non_meta_recommender(batch_size, pending_experiments)
        if isinstance(recommender, BayesianRecommender):
            return recommender.get_surrogate(
                self.searchspace, self.objective, self.measurements
            )
        else:
            raise IncompatibilityError(
                f"The current recommender is of type "
                f"'{recommender.__class__.__name__}', which does not provide "
                f"a surrogate model. Surrogate models are only available for "
                f"recommender subclasses of '{BayesianRecommender.__name__}'."
            )

    def _get_non_meta_recommender(
        self,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> RecommenderProtocol:
        """Get the current recommender.

        Args:
            batch_size: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.
            pending_experiments: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.

        Returns:
            The recommender for the current recommendation context.
        """
        if not isinstance(self.recommender, MetaRecommender):
            return self.recommender
        return self.recommender.get_non_meta_recommender(
            batch_size,
            self.searchspace,
            self.objective,
            self.measurements,
            pending_experiments,
        )

    def _get_bayesian_recommender(
        self,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> BayesianRecommender:
        """Get the current Bayesian recommender (if available).

        For details on the method arguments, see :meth:`_get_non_meta_recommender`.
        """
        recommender = self._get_non_meta_recommender(batch_size, pending_experiments)
        if not isinstance(recommender, BayesianRecommender):
            raise IncompatibilityError(
                f"The current recommender is of type "
                f"'{recommender.__class__.__name__}', which does not provide "
                f"a surrogate model or acquisition values. Both objects are "
                f"only available for recommender subclasses of "
                f"'{BayesianRecommender.__name__}'."
            )
        return recommender

    def get_acquisition_function(
        self,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> AcquisitionFunction:
        """Get the current acquisition function.

        Args:
            batch_size: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.
            pending_experiments: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.

        Raises:
            IncompatibilityError: If no objective has been specified.
            IncompatibilityError: If the current recommender does not use an acquisition
                function.

        Returns:
            The acquisition function of the current recommender.
        """
        if self.objective is None:
            raise IncompatibilityError(
                "Acquisition values can only be computed if an objective has "
                "been defined."
            )

        recommender = self._get_bayesian_recommender(batch_size, pending_experiments)
        return recommender.get_acquisition_function(
            self.searchspace, self.objective, self.measurements, pending_experiments
        )

    def acquisition_values(
        self,
        candidates: pd.DataFrame,
        acquisition_function: AcquisitionFunction | None = None,
        *,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.Series:
        """Compute the acquisition values for the given candidates.

        Args:
            candidates: The candidate points in experimental representation.
                For details, see :meth:`baybe.surrogates.base.Surrogate.posterior`.
            acquisition_function: The acquisition function to be evaluated.
                If not provided, the acquisition function of the recommender is used.
            batch_size: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.
            pending_experiments: See :meth:`recommend`.
                Only required when using meta recommenders that demand it.

        Returns:
            A series of individual acquisition values, one for each candidate.
        """
        recommender = self._get_bayesian_recommender(batch_size, pending_experiments)
        assert self.objective is not None
        return recommender.acquisition_values(
            candidates,
            self.searchspace,
            self.objective,
            self.measurements,
            pending_experiments,
            acquisition_function,
        )

    def joint_acquisition_value(  # noqa: DOC101, DOC103
        self,
        candidates: pd.DataFrame,
        acquisition_function: AcquisitionFunction | None = None,
        *,
        batch_size: int | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> float:
        """Compute the joint acquisition values for the given candidate batch.

        For details on the method arguments, see :meth:`acquisition_values`.

        Returns:
            The joint acquisition value of the batch.
        """
        recommender = self._get_bayesian_recommender(batch_size, pending_experiments)
        assert self.objective is not None
        return recommender.joint_acquisition_value(
            candidates,
            self.searchspace,
            self.objective,
            self.measurements,
            pending_experiments,
            acquisition_function,
        )


def _add_version(dict_: dict) -> dict:
    """Add the package version to the given dictionary."""
    from baybe import __version__

    return {**dict_, "version": __version__}


def _drop_version(dict_: dict) -> dict:
    """Drop the package version from the given dictionary."""
    dict_.pop("version", None)
    return dict_


# Register (un-)structure hooks
unstructure_hook = cattrs.gen.make_dict_unstructure_fn(
    Campaign, converter, _cattrs_include_init_false=True
)
structure_hook = cattrs.gen.make_dict_structure_fn(
    Campaign, converter, _cattrs_include_init_false=True, _cattrs_forbid_extra_keys=True
)
converter.register_unstructure_hook(
    Campaign, lambda x: _add_version(unstructure_hook(x))
)
converter.register_structure_hook(
    Campaign, lambda d, cl: structure_hook(_drop_version(d), cl)
)


# Converter for config validation
_validation_converter = converter.copy()
_validation_converter.register_structure_hook(
    SearchSpace, validate_searchspace_from_config
)

# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Transfer learning in a continuous space with negated Michalewicz functions as tasks.

This uses the 5-dimensional version of the Michalewictz function and adds
a noise_std of 0.15 to the source function.
"""

from __future__ import annotations

import math
from collections.abc import Callable

import numpy as np
import pandas as pd
import torch
from botorch.test_functions.synthetic import Michalewicz

from baybe.campaign import Campaign
from baybe.objectives import SingleTargetObjective
from baybe.parameters import NumericalContinuousParameter, TaskParameter
from baybe.parameters.base import Parameter
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from benchmarks.definition import ConvergenceBenchmark, ConvergenceBenchmarkSettings


def make_searchspace(use_task_parameter: bool) -> SearchSpace:
    """Create search space for the benchmark."""
    params: list[Parameter] = [
        NumericalContinuousParameter(
            name=f"x{k}",
            bounds=(0, math.pi),
        )
        for k in range(5)
    ]
    if use_task_parameter:
        params.append(
            TaskParameter(
                name="Function",
                values=["Target_Function", "Source_Function"],
                active_values=["Target_Function"],
            )
        )

    return SearchSpace.from_product(parameters=params)


def make_objective() -> SingleTargetObjective:
    """Create the objective for the benchmark."""
    return SingleTargetObjective(target=NumericalTarget(name="Target", mode="MAX"))


def wrap_function(
    function: Callable, function_name: str, df: pd.DataFrame
) -> pd.DataFrame:
    """Wrap the given function such that it operates on DataFrames."""
    input_columns = df.columns.tolist()

    # Create a copy to avoid modifying the original DataFrame
    result_df = df.copy()

    # Apply Michalewicz function to each row
    def apply_function(row):
        # Remove the "Function" column but record its value such that it can be added
        # back to the row after the function evaluation
        if "Function" in row.index:
            function_value = row["Function"]
            row = row.drop("Function")
        else:
            function_value = None  # Default if column doesn't exist

        x = torch.tensor(row.values.astype(float), dtype=torch.float64).unsqueeze(0)
        # Re-add the "Function" column to the row
        if function_value is not None:
            row["Function"] = function_value
        return function(x).item()

    result_df["Target"] = (
        result_df[input_columns].apply(apply_function, axis=1).astype(float)
    )
    # Add a column "Function" with the function name
    result_df["Function"] = function_name

    return result_df


def make_initial_data(
    function: Callable, function_name: str, num_of_points: int
) -> pd.DataFrame:
    """Create initial data points for the Michalewicz benchmark."""
    # Create random samples in [0, pi]^dim
    samples = np.random.uniform(low=0, high=math.pi, size=(num_of_points, 5))

    # Convert to DataFrame
    column_names = [f"x{i}" for i in range(5)]
    df = pd.DataFrame(samples, columns=column_names)

    # Apply the function to get target values
    df = wrap_function(function, function_name, df)

    return df


def michalewicz_tl_continuous(settings: ConvergenceBenchmarkSettings) -> pd.DataFrame:
    """Benchmark function for transfer learning with the Michalewicz function in 5D.

    Key characteristics:
    • Compares transfer learning vs. non-transfer learning approaches
    • Uses two versions of Michalewicz function:
      - Target: standard negated Michalewicz
      - Source: negated Michalewicz with added noise (noise_std=0.15)
    • Tests varying amounts of source data (1, 10, 50, 100 points)
    • Includes baseline with no transfer learning (0 points)
    • Creates two campaign types:
      - Transfer learning: with task parameter to distinguish data sources
      - Non-transfer learning (naive): without task parameter

    Args:
        settings: Configuration settings for the convergence benchmark

    Returns:
        DataFrame containing benchmark results for all test cases
    """
    functions = {
        "Source_Function": Michalewicz(dim=5, negate=True, noise_std=0.15),
        "Target_Function": Michalewicz(dim=5, negate=True),
    }
    searchspace_nontl = make_searchspace(use_task_parameter=False)
    searchspace_tl = make_searchspace(use_task_parameter=True)

    objective = make_objective()
    campaign_tl = Campaign(
        searchspace=searchspace_tl,
        objective=objective,
    )
    campaign_nontl = Campaign(
        searchspace=searchspace_nontl,
        objective=objective,
    )

    results = []

    for p in [1, 10, 50, 100]:
        results.append(
            simulate_scenarios(
                {f"{p}": campaign_tl, f"{p}_naive": campaign_nontl},
                lambda x: wrap_function(
                    functions["Target_Function"], "Target_Function", x
                ),
                initial_data=[
                    make_initial_data(
                        functions["Source_Function"], "Source_Function", p
                    )
                    for _ in range(settings.n_mc_iterations)
                ],
                batch_size=settings.batch_size,
                n_doe_iterations=settings.n_doe_iterations,
                impute_mode="error",
            )
        )
    results.append(
        simulate_scenarios(
            {"0": campaign_tl, "0_naive": campaign_nontl},
            lambda x: wrap_function(functions["Target_Function"], "Target_Function", x),
            batch_size=settings.batch_size,
            n_doe_iterations=settings.n_doe_iterations,
            n_mc_iterations=settings.n_mc_iterations,
            impute_mode="error",
        )
    )
    return pd.concat(results)


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=2,
    n_doe_iterations=30,
    n_mc_iterations=30,
)

michalewicz_tl_continuous_benchmark = ConvergenceBenchmark(
    function=michalewicz_tl_continuous,
    optimal_target_values={"Target": 4.687658},
    settings=benchmark_config,
)



"""Convergence benchmark configuration."""

from typing import Any

from attrs import define, field
from attrs.validators import deep_mapping, instance_of, optional

from benchmarks.definition.base import Benchmark, BenchmarkSettings


@define(frozen=True, kw_only=True)
class ConvergenceBenchmarkSettings(BenchmarkSettings):
    """Benchmark configuration for recommender convergence analyses."""

    batch_size: int = field(validator=instance_of(int))
    """The recommendation batch size."""

    n_doe_iterations: int = field(validator=instance_of(int))
    """The number of Design of Experiment iterations."""

    n_mc_iterations: int = field(validator=instance_of(int))
    """The number of Monte Carlo iterations."""


@define(frozen=True)
class ConvergenceBenchmark(Benchmark[ConvergenceBenchmarkSettings]):
    """A class for defining convergence benchmarks."""

    optimal_target_values: dict[str, Any] | None = field(
        default=None,
        validator=optional(
            deep_mapping(
                key_validator=instance_of(str),
                mapping_validator=instance_of(dict),
                value_validator=lambda *_: None,
            )
        ),
    )
    """The optimal values that can be achieved for the targets **individually**."""



"""Utilities for working with surrogates."""

from __future__ import annotations

from collections.abc import Callable
from functools import wraps
from typing import TYPE_CHECKING, TypeVar

from baybe.exceptions import InvalidSurrogateModelError
from baybe.surrogates.base import Surrogate

if TYPE_CHECKING:
    from botorch.posteriors import Posterior
    from torch import Tensor

    from baybe.surrogates.naive import MeanPredictionSurrogate

_TSurrogate = TypeVar("_TSurrogate", bound=Surrogate)


_constant_target_model_store: dict[int, MeanPredictionSurrogate] = {}
"""Dictionary for storing constant target fallback models. Keys are the IDs of the
surrogate models that temporarily have a fallback attached because they were
trained on constant training targets. Values are the corresponding fallback models."""


def catch_constant_targets(
    cls: type[_TSurrogate], std_threshold: float = 1e-6
) -> type[_TSurrogate]:
    """Make a ``Surrogate`` class robustly handle constant training targets.

    More specifically, "constant training targets" can mean either of:
        * The standard deviation of the training targets is below the given threshold.
        * There is only one target and the standard deviation cannot even be computed.

    The modified class handles the above cases separately from "regular operation"
    by resorting to a :class:`baybe.surrogates.naive.MeanPredictionSurrogate`,
    which is stored outside the model in a dictionary maintained by this decorator.

    Args:
        cls: The :class:`baybe.surrogates.base.Surrogate` to be augmented.
        std_threshold: The standard deviation threshold below which operation is
            switched to the alternative model.

    Returns:
        The modified class.
    """
    from baybe.surrogates.naive import MeanPredictionSurrogate

    # References to original methods
    _fit_original = cls._fit
    _posterior_original = cls._posterior

    def _posterior_new(self, candidates_comp_scaled: Tensor, /) -> Posterior:
        """Use fallback model if it exists, otherwise call original posterior."""
        # Alternative model fallback
        if constant_target_model := _constant_target_model_store.get(id(self), None):
            return constant_target_model._posterior(candidates_comp_scaled)

        # Regular operation
        return _posterior_original(self, candidates_comp_scaled)

    def _fit_new(self, train_x: Tensor, train_y: Tensor) -> None:
        """Original fit but with fallback model creation for constant targets."""
        if not (train_y.ndim == 2 and train_y.shape[-1] == 1):
            raise NotImplementedError(
                "The current logic is only implemented for single-target surrogates."
            )

        # Alternative model fallback
        if train_y.numel() == 1 or train_y.std() < std_threshold:
            model = MeanPredictionSurrogate()
            model._fit(train_x, train_y)
            _constant_target_model_store[id(self)] = model

        # Regular operation
        else:
            _constant_target_model_store.pop(id(self), None)
            _fit_original(self, train_x, train_y)

    # Replace the methods
    cls._posterior = _posterior_new  # type: ignore
    cls._fit = _fit_new  # type: ignore

    return cls


def batchify_mean_var_prediction(
    posterior: Callable[[_TSurrogate, Tensor], tuple[Tensor, Tensor]],
) -> Callable[[_TSurrogate, Tensor], tuple[Tensor, Tensor]]:
    """Wrap a posterior method to make it evaluate t-batches as an augmented q-batch."""

    @wraps(posterior)
    def sequential_posterior(
        model: _TSurrogate, candidates: Tensor
    ) -> tuple[Tensor, Tensor]:
        # If no batch dimensions are given, call the model directly
        if candidates.ndim == 2:
            return posterior(model, candidates)

        # Parameter batching is not (yet) supported
        if candidates.ndim > 3:
            raise ValueError("Multiple t-batch dimensions are not supported.")

        # Keep track of batch dimensions
        t_shape = candidates.shape[-3]
        q_shape = candidates.shape[-2]

        # Flatten the t-batch dimension into the q-batch dimension
        flattened = candidates.flatten(end_dim=-2)

        # Call the model on the entire input
        mean, var = posterior(model, flattened)

        # Restore the batch dimensions
        mean = mean.reshape((t_shape, q_shape))
        var = var.reshape((t_shape, q_shape))

        return mean, var

    return sequential_posterior


def batchify_ensemble_predictor(
    base_predictor: Callable[[Tensor], Tensor],
) -> Callable[[Tensor], Tensor]:
    """Wrap an ensemble predictor to make it evaluate t-batches as an augmented q-batch.

    Args:
        base_predictor: The ensemble predictor to be wrapped.

    Returns:
        The wrapped predictor.
    """

    @wraps(base_predictor)
    def batch_predictor(candidates: Tensor) -> Tensor:
        # If no batch dimensions are given, call the model directly
        if candidates.ndim == 2:
            return base_predictor(candidates)

        # Ensemble models do not (yet) support model parameter batching
        if candidates.ndim > 3:
            raise ValueError("Multiple t-batch dimensions are not supported.")

        # Keep track of batch dimensions
        t_shape = candidates.shape[-3]
        q_shape = candidates.shape[-2]

        # Flatten the t-batch dimension into the q-batch dimension
        flattened = candidates.flatten(end_dim=-2)

        # Call the model on the entire input
        predictions = base_predictor(flattened)

        # Assert that the model provides the ensemble predictions in the correct shape
        # (otherwise the reshaping operation below could silently produce wrong results)
        try:
            assert predictions.ndim == 2
            n_estimators = predictions.shape[0]
            assert predictions.shape[1] == t_shape * q_shape
        except AssertionError:
            raise InvalidSurrogateModelError(
                f"For the given input of shape {tuple(candidates.shape)}, "
                f"the ensemble model is supposed to create predictions of shape "
                f"(n_estimators, t_shape * q_shape) = "
                f"(n_estimators, {t_shape * q_shape}) "
                f"but returned an array of incompatible shape "
                f"{tuple(predictions.shape)}."
            )

        # Restore the batch dimensions
        predictions = predictions.reshape((n_estimators, t_shape, q_shape))
        predictions = predictions.permute((1, 0, 2))

        return predictions

    return batch_predictor



"""Hartmann function in 3 dimensions as a benchmark."""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np
from botorch.test_functions.synthetic import Hartmann
from pandas import DataFrame

from baybe.campaign import Campaign
from baybe.parameters import NumericalContinuousParameter
from baybe.recommenders import RandomRecommender
from baybe.searchspace import SearchSpace
from baybe.simulation import simulate_scenarios
from baybe.targets import NumericalTarget
from baybe.utils.dataframe import arrays_to_dataframes
from benchmarks.definition.convergence import (
    ConvergenceBenchmark,
    ConvergenceBenchmarkSettings,
)

if TYPE_CHECKING:
    from mpl_toolkits.mplot3d import Axes3D


def hartmann_3d(settings: ConvergenceBenchmarkSettings) -> DataFrame:
    """Benchmark function with the Hartmann 3D test function.

    Key characteristics:
    • Parameters:
      - x1: Continuous [0, 1]
      - x2: Continuous [0, 1]
      - x3: Continuous [0, 1]
    • Output: Continuous
    • Objective: Minimization
    • Optimal inputs: {x1: 0.114614, x2: 0.555649, x3: 0.852547}
    • Optimal output: -3.86278
    • Tests multiple recommenders:
      - Random Recommender
      - Default Recommender

    Args:
        settings: Configuration settings for the convergence benchmark.

    Returns:
        DataFrame containing benchmark results.
    """
    parameters = [
        NumericalContinuousParameter("x1", (0.0, 1.0)),
        NumericalContinuousParameter("x2", (0.0, 1.0)),
        NumericalContinuousParameter("x3", (0.0, 1.0)),
    ]

    target = NumericalTarget(name="target", mode="MIN")
    searchspace = SearchSpace.from_product(parameters=parameters)
    objective = target.to_objective()

    scenarios: dict[str, Campaign] = {
        "Random Recommender": Campaign(
            searchspace=searchspace,
            recommender=RandomRecommender(),
            objective=objective,
        ),
        "Default Recommender": Campaign(
            searchspace=searchspace,
            objective=objective,
        ),
    }

    hartmann = Hartmann(dim=3)

    lookup = arrays_to_dataframes(
        [p.name for p in parameters], [target.name], use_torch=True
    )(hartmann)

    return simulate_scenarios(
        scenarios,
        lookup,
        batch_size=settings.batch_size,
        n_doe_iterations=settings.n_doe_iterations,
        n_mc_iterations=settings.n_mc_iterations,
        impute_mode="error",
    )


benchmark_config = ConvergenceBenchmarkSettings(
    batch_size=5,
    n_doe_iterations=30,
    n_mc_iterations=100,
)

hartmann_3d_benchmark = ConvergenceBenchmark(
    function=hartmann_3d,
    optimal_target_values={"target": -3.86278},
    settings=benchmark_config,
)


if __name__ == "__main__":
    # Visualize the domain (limited to 2D slices due to 3D nature)

    import matplotlib.pyplot as plt

    # Create a grid of points for x1 and x2 dimensions
    x1 = np.linspace(0, 1, 50)
    x2 = np.linspace(0, 1, 50)
    x1_mesh, x2_mesh = np.meshgrid(x1, x2)

    # Create slices for different fixed values of x3
    x3_values = [0.1, 0.3, 0.5, 0.7, 0.9]

    # Instantiate the Hartmann function
    hartmann_func = Hartmann(dim=3, negate=False)

    fig = plt.figure(figsize=(15, 10))
    for i, x3_val in enumerate(x3_values):
        ax: Axes3D = fig.add_subplot(2, 3, i + 1, projection="3d")
        x3_mesh = np.full_like(x1_mesh, x3_val)

        # Stack the meshgrid points into the required shape
        points = np.column_stack((x1_mesh.ravel(), x2_mesh.ravel(), x3_mesh.ravel()))

        # Calculate function values
        import torch

        with torch.no_grad():
            values = (
                hartmann_func(torch.tensor(points, dtype=torch.float64))
                .numpy()
                .reshape(x1_mesh.shape)
            )

        # Plot the surface
        surf = ax.plot_surface(x1_mesh, x2_mesh, values, cmap="viridis")

        ax.set_xlabel("x1")
        ax.set_ylabel("x2")
        ax.set_zlabel("Hartmann-3D")
        ax.set_title(f"x3 = {x3_val}")

    plt.tight_layout()
    plt.show()



"""Base class for all Bayesian recommenders."""

from __future__ import annotations

import gc
import warnings
from abc import ABC
from typing import TYPE_CHECKING

import pandas as pd
from attrs import define, field, fields
from attrs.converters import optional
from typing_extensions import override

from baybe.acquisition import qLogEI, qLogNEHVI
from baybe.acquisition.base import AcquisitionFunction
from baybe.acquisition.utils import convert_acqf
from baybe.exceptions import (
    IncompatibleAcquisitionFunctionError,
    InvalidSurrogateModelError,
)
from baybe.objectives.base import Objective
from baybe.recommenders.pure.base import PureRecommender
from baybe.searchspace import SearchSpace
from baybe.surrogates import CustomONNXSurrogate, GaussianProcessSurrogate
from baybe.surrogates.base import (
    IndependentGaussianSurrogate,
    Surrogate,
    SurrogateProtocol,
)
from baybe.utils.dataframe import _ValidatedDataFrame, normalize_input_dtypes
from baybe.utils.validation import validate_parameter_input, validate_target_input

if TYPE_CHECKING:
    from botorch.acquisition import AcquisitionFunction as BoAcquisitionFunction


def _autoreplicate(surrogate: SurrogateProtocol, /) -> SurrogateProtocol:
    """Replicates single-output surrogate models and passes through everything else."""
    if isinstance(surrogate, Surrogate) and not surrogate.supports_multi_output:
        return surrogate.replicate()
    return surrogate


@define
class BayesianRecommender(PureRecommender, ABC):
    """An abstract class for Bayesian Recommenders."""

    _surrogate_model: SurrogateProtocol = field(
        alias="surrogate_model", factory=GaussianProcessSurrogate
    )
    """The surrogate model."""

    acquisition_function: AcquisitionFunction | None = field(
        default=None, converter=optional(convert_acqf)
    )
    """The acquisition function. When omitted, a default is used."""

    # TODO: The objective is currently only required for validating the recommendation
    #   context. Once multi-target support is complete, we might want to refactor
    #   the validation mechanism, e.g. by
    #   * storing only the minimal low-level information required
    #   * switching to a strategy where we catch the BoTorch exceptions
    #   * ...
    _objective: Objective | None = field(default=None, init=False, eq=False)
    """The encountered objective to be optimized."""

    _botorch_acqf = field(default=None, init=False, eq=False)
    """The induced BoTorch acquisition function."""

    @property
    def surrogate_model(self) -> SurrogateProtocol:
        """Deprecated!"""
        warnings.warn(
            f"Accessing the surrogate model via 'surrogate_model' has been "
            f"deprecated. Use '{self.get_surrogate.__name__}' instead to get the "
            f"trained model instance (or "
            f"'{fields(type(self))._surrogate_model.name}' to access the raw object).",
            DeprecationWarning,
        )
        return self._surrogate_model

    def _get_acquisition_function(self, objective: Objective) -> AcquisitionFunction:
        """Select the appropriate default acquisition function for the given context."""
        if self.acquisition_function is None:
            return qLogNEHVI() if objective.is_multi_output else qLogEI()
        return self.acquisition_function

    def get_surrogate(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
    ) -> SurrogateProtocol:
        """Get the trained surrogate model."""
        # This fit applies internal caching and does not necessarily involve computation
        surrogate = (
            _autoreplicate(self._surrogate_model)
            if objective.is_multi_output
            else self._surrogate_model
        )
        surrogate.fit(searchspace, objective, measurements)

        return surrogate

    def _setup_botorch_acqf(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
    ) -> None:
        """Create the acquisition function for the current training data."""  # noqa: E501
        self._objective = objective
        acqf = self._get_acquisition_function(objective)

        if objective.is_multi_output and not acqf.supports_multi_output:
            raise IncompatibleAcquisitionFunctionError(
                f"You attempted to use a single-output acquisition function in a "
                f"{len(objective.targets)}-target multi-output context."
            )

        surrogate = self.get_surrogate(searchspace, objective, measurements)
        self._botorch_acqf = acqf.to_botorch(
            surrogate,
            searchspace,
            objective,
            measurements,
            pending_experiments,
        )

    def get_acquisition_function(
        self,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
    ) -> BoAcquisitionFunction:
        """Get the acquisition function for the given recommendation context.

        For details on the method arguments, see :meth:`recommend`.
        """
        self._setup_botorch_acqf(
            searchspace, objective, measurements, pending_experiments
        )
        return self._botorch_acqf

    @override
    def recommend(
        self,
        batch_size: int,
        searchspace: SearchSpace,
        objective: Objective | None = None,
        measurements: pd.DataFrame | None = None,
        pending_experiments: pd.DataFrame | None = None,
    ) -> pd.DataFrame:
        if objective is None:
            raise NotImplementedError(
                f"Recommenders of type '{BayesianRecommender.__name__}' require "
                f"that an objective is specified."
            )

        # Experimental input validation
        if (measurements is None) or measurements.empty:
            raise NotImplementedError(
                f"Recommenders of type '{BayesianRecommender.__name__}' do not support "
                f"empty training data."
            )
        if not isinstance(measurements, _ValidatedDataFrame):
            validate_target_input(measurements, objective.targets)
            validate_parameter_input(measurements, searchspace.parameters)
            measurements = normalize_input_dtypes(
                measurements, searchspace.parameters + objective.targets
            )
            measurements.__class__ = _ValidatedDataFrame
        if pending_experiments is not None and not isinstance(
            pending_experiments, _ValidatedDataFrame
        ):
            validate_parameter_input(pending_experiments, searchspace.parameters)
            pending_experiments = normalize_input_dtypes(
                pending_experiments, searchspace.parameters
            )
            pending_experiments.__class__ = _ValidatedDataFrame

        if (
            isinstance(self._surrogate_model, IndependentGaussianSurrogate)
            and batch_size > 1
        ):
            raise InvalidSurrogateModelError(
                f"The specified surrogate model of type "
                f"'{self._surrogate_model.__class__.__name__}' "
                f"cannot be used for batch recommendation."
            )

        if isinstance(self._surrogate_model, CustomONNXSurrogate):
            CustomONNXSurrogate.validate_compatibility(searchspace)

        self._setup_botorch_acqf(
            searchspace, objective, measurements, pending_experiments
        )

        return super().recommend(
            batch_size=batch_size,
            searchspace=searchspace,
            objective=objective,
            measurements=measurements,
            pending_experiments=pending_experiments,
        )

    def acquisition_values(
        self,
        candidates: pd.DataFrame,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
        acquisition_function: AcquisitionFunction | None = None,
    ) -> pd.Series:
        """Compute the acquisition values for the given candidates.

        Args:
            candidates: The candidate points in experimental representation.
                For details, see :meth:`baybe.surrogates.base.Surrogate.posterior`.
            searchspace:
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            objective:
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            measurements:
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            pending_experiments:
                See :meth:`baybe.recommenders.base.RecommenderProtocol.recommend`.
            acquisition_function: The acquisition function to be evaluated.
                If not provided, the acquisition function of the recommender is used.

        Returns:
            A series of individual acquisition values, one for each candidate.
        """
        surrogate = self.get_surrogate(searchspace, objective, measurements)
        acqf = acquisition_function or self._get_acquisition_function(objective)
        return acqf.evaluate(
            candidates,
            surrogate,
            searchspace,
            objective,
            measurements,
            pending_experiments,
            jointly=False,
        )

    def joint_acquisition_value(  # noqa: DOC101, DOC103
        self,
        candidates: pd.DataFrame,
        searchspace: SearchSpace,
        objective: Objective,
        measurements: pd.DataFrame,
        pending_experiments: pd.DataFrame | None = None,
        acquisition_function: AcquisitionFunction | None = None,
    ) -> float:
        """Compute the joint acquisition value for the given candidate batch.

        For details on the method arguments, see :meth:`acquisition_values`.

        Returns:
            The joint acquisition value of the batch.
        """
        surrogate = self.get_surrogate(searchspace, objective, measurements)
        acqf = acquisition_function or self._get_acquisition_function(objective)
        return acqf.evaluate(
            candidates,
            surrogate,
            searchspace,
            objective,
            measurements,
            pending_experiments,
            jointly=True,
        )


# Collect leftover original slotted classes processed by `attrs.define`
gc.collect()



"""Tests features of the Campaign object."""

from contextlib import nullcontext

import pandas as pd
import pytest
from pandas.testing import assert_index_equal
from pytest import param

from baybe.acquisition import qLogEI, qLogNEHVI, qTS, qUCB
from baybe.campaign import _EXCLUDED, Campaign
from baybe.constraints.conditions import SubSelectionCondition
from baybe.constraints.discrete import DiscreteExcludeConstraint
from baybe.objectives import DesirabilityObjective, ParetoObjective
from baybe.parameters.numerical import (
    NumericalContinuousParameter,
    NumericalDiscreteParameter,
)
from baybe.searchspace.core import SearchSpaceType
from baybe.searchspace.discrete import SubspaceDiscrete
from baybe.surrogates import (
    BetaBernoulliMultiArmedBanditSurrogate,
    GaussianProcessSurrogate,
)
from baybe.targets import BinaryTarget, NumericalTarget
from baybe.utils.basic import UNSPECIFIED

from .conftest import run_iterations


@pytest.mark.parametrize(
    "target_names",
    [
        param(["Target_max"], id="max"),
        param(["Target_min"], id="min"),
        param(["Target_max_bounded"], id="max_b"),
        param(["Target_min_bounded"], id="min_b"),
        param(["Target_match_bell"], id="match_bell"),
        param(["Target_match_triangular"], id="match_tri"),
        param(
            ["Target_max_bounded", "Target_min_bounded", "Target_match_triangular"],
            id="desirability",
        ),
    ],
)
@pytest.mark.parametrize("batch_size", [2], ids=["b2"])
@pytest.mark.parametrize("n_iterations", [2], ids=["i2"])
def test_get_surrogate(campaign, n_iterations, batch_size):
    """Test successful extraction of the surrogate model."""
    run_iterations(campaign, n_iterations, batch_size)

    model = campaign.get_surrogate()
    assert model is not None, "Something went wrong during surrogate model extraction."


@pytest.mark.parametrize("complement", [False, True], ids=["regular", "complement"])
@pytest.mark.parametrize("exclude", [True, False], ids=["exclude", "include"])
@pytest.mark.parametrize(
    "constraints",
    [
        pd.DataFrame({"a": [0]}),
        [DiscreteExcludeConstraint(["a"], [SubSelectionCondition([1])])],
    ],
    ids=["dataframe", "constraints"],
)
def test_candidate_toggling(constraints, exclude, complement):
    """Toggling discrete candidates updates the campaign metadata accordingly."""
    subspace = SubspaceDiscrete.from_product(
        [
            NumericalDiscreteParameter("a", [0, 1]),
            NumericalDiscreteParameter("b", [3, 4, 5]),
        ]
    )
    campaign = Campaign(subspace)

    # Set metadata to opposite of targeted value so that we can verify the effect later
    campaign._searchspace_metadata[_EXCLUDED] = not exclude

    # Toggle the candidates
    campaign.toggle_discrete_candidates(constraints, exclude, complement=complement)

    # Extract row indices of candidates whose metadata should have been toggled
    matches = campaign.searchspace.discrete.exp_rep["a"] == 0
    idx = matches.index[~matches if complement else matches]

    # Assert that metadata is set correctly
    target = campaign._searchspace_metadata.loc[idx, _EXCLUDED]
    other = campaign._searchspace_metadata[_EXCLUDED].drop(index=idx)
    assert all(target == exclude)  # must contain the updated values
    assert all(other != exclude)  # must contain the original values


@pytest.mark.parametrize(
    "flag",
    [
        "allow_recommending_already_measured",
        "allow_recommending_already_recommended",
        "allow_recommending_pending_experiments",
    ],
    ids=lambda x: x.removeprefix("allow_recommending_"),
)
@pytest.mark.parametrize(
    "space_type",
    [SearchSpaceType.DISCRETE, SearchSpaceType.CONTINUOUS],
    ids=lambda x: x.name,
)
@pytest.mark.parametrize(
    "value", [True, False, param(UNSPECIFIED, id=repr(UNSPECIFIED))]
)
def test_setting_allow_flags(flag, space_type, value):
    """Passed allow_* flags are rejected if incompatible with the search space type."""
    kwargs = {flag: value}
    expect_error = (space_type is SearchSpaceType.DISCRETE) != (
        value is not UNSPECIFIED
    )

    if space_type is SearchSpaceType.DISCRETE:
        parameter = NumericalDiscreteParameter("p", [0, 1])
    else:
        parameter = NumericalContinuousParameter("p", [0, 1])

    with pytest.raises(ValueError) if expect_error else nullcontext():
        Campaign(parameter, **kwargs)


@pytest.mark.parametrize(
    "parameter_names", [["Categorical_1", "Categorical_2", "Num_disc_1"]]
)
@pytest.mark.parametrize("n_iterations", [3], ids=["i3"])
def test_update_measurements(ongoing_campaign):
    """Updating measurements makes the expected changes."""
    p_name, t_name = "Num_disc_1", ongoing_campaign.targets[0].name
    updated = ongoing_campaign.measurements.iloc[[0], :]

    # Perform the update
    updated.iloc[0, updated.columns.get_loc(p_name)] = 1337
    updated.iloc[0, updated.columns.get_loc(t_name)] = 1337
    ongoing_campaign.update_measurements(
        updated, numerical_measurements_must_be_within_tolerance=False
    )

    # Make sure values are updated and resets have been made
    meas = ongoing_campaign.measurements
    assert meas.iloc[0, updated.columns.get_loc(p_name)] == 1337
    assert meas.iloc[0, updated.columns.get_loc(t_name)] == 1337
    assert meas.iloc[[0], updated.columns.get_loc("FitNr")].isna().all()
    assert ongoing_campaign._cached_recommendation.empty


@pytest.mark.parametrize(
    ("parameter_names", "objective", "surrogate_model", "acqf", "batch_size"),
    [
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            NumericalTarget("t1", "MAX").to_objective(),
            GaussianProcessSurrogate(),
            qLogEI(),
            3,
            id="single_target",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            DesirabilityObjective(
                (
                    NumericalTarget("t1", "MAX", bounds=(0, 1)),
                    NumericalTarget("t2", "MIN", bounds=(0, 1)),
                )
            ),
            GaussianProcessSurrogate(),
            qLogEI(),
            3,
            id="desirability",
        ),
        param(
            ["Categorical_1", "Num_disc_1", "Conti_finite1"],
            ParetoObjective(
                (NumericalTarget("t1", "MAX"), NumericalTarget("t2", "MIN"))
            ),
            GaussianProcessSurrogate(),
            qLogNEHVI(),
            3,
            id="pareto",
        ),
        param(
            ["Categorical_1"],
            BinaryTarget(name="Target_binary").to_objective(),
            BetaBernoulliMultiArmedBanditSurrogate(),
            qTS(),
            1,
            id="bernoulli",
        ),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("n_iterations", [1], ids=["i1"])
def test_posterior_stats(ongoing_campaign, n_iterations, batch_size):
    """Posterior statistics have expected shape, index and columns."""
    objective = ongoing_campaign.objective
    tested_stats = ["mean", "std", "var"]
    test_quantiles = not isinstance(objective.targets[0], BinaryTarget)
    if test_quantiles:
        tested_stats += [0.05, 0.95]

    stats = ongoing_campaign.posterior_stats(
        ongoing_campaign.measurements, tested_stats
    )

    # Assert number of entries and index
    (
        assert_index_equal(ongoing_campaign.measurements.index, stats.index),
        (ongoing_campaign.measurements.index, stats.index),
    )

    # Assert expected columns are present.
    match objective:
        case DesirabilityObjective():
            targets = ["Desirability"]
        case _:
            targets = [t.name for t in objective.targets]

    for t in targets:
        for stat in tested_stats:
            stat_name = f"Q_{stat}" if isinstance(stat, float) else stat
            assert sum(f"{t}_{stat_name}" in x for x in stats.columns) == 1, (
                f"{t}_{stat_name} not in the returned posterior statistics"
            )

    # Assert no NaN's present
    assert not stats.isna().any().any()

    if not test_quantiles:
        # Assert correct error for unsupported quantile calculation
        with pytest.raises(
            TypeError, match="does not support the statistic associated"
        ):
            ongoing_campaign.posterior_stats(ongoing_campaign.measurements, [0.1])


@pytest.mark.parametrize(
    ("stats", "error", "match"),
    [
        param(
            ["invalid"],
            TypeError,
            "does not support the statistic associated",
            id="invalid_stat",
        ),
        param(
            [-0.1],
            ValueError,
            "quantile statistics can only be",
            id="quantile_too_small",
        ),
        param(
            [1.1],
            ValueError,
            "quantile statistics can only be",
            id="quantile_too_large",
        ),
    ],
)
@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("n_iterations", [1], ids=["i1"])
@pytest.mark.parametrize("batch_size", [1], ids=["b3"])
def test_posterior_stats_invalid_input(ongoing_campaign, stats, error, match):
    """Invalid inputs for posterior statistics raise expected exceptions."""
    with pytest.raises(error, match=match):
        ongoing_campaign.posterior_stats(ongoing_campaign.measurements, stats)


@pytest.mark.parametrize("n_grid_points", [5], ids=["g5"])
@pytest.mark.parametrize("n_iterations", [1], ids=["i1"])
@pytest.mark.parametrize("batch_size", [3], ids=["b3"])
def test_acquisition_value_computation(ongoing_campaign: Campaign):
    """Acquisition values have the expected shape."""
    df = ongoing_campaign.searchspace.discrete.exp_rep
    assert not df.empty

    # Using campaign acquisition function
    acqfs = ongoing_campaign.acquisition_values(df)
    assert_index_equal(acqfs.index, df.index)
    joint_acqf = ongoing_campaign.joint_acquisition_value(df)
    assert isinstance(joint_acqf, float)

    # Using separate acquisition function
    acqfs = ongoing_campaign.acquisition_values(df, qUCB())
    assert_index_equal(acqfs.index, df.index)
    joint_acqf = ongoing_campaign.joint_acquisition_value(df, qUCB())
    assert isinstance(joint_acqf, float)



"""Tests for sampling algorithm utilities."""

import math

import hypothesis.extra.numpy as hnp
import hypothesis.strategies as st
import numpy as np
import pandas as pd
import pytest
from hypothesis import example, given
from sklearn.metrics import pairwise_distances

from baybe.utils.sampling_algorithms import (
    DiscreteSamplingMethod,
    farthest_point_sampling,
    sample_numerical_df,
)


@pytest.mark.parametrize("fraction", [0.2, 0.8, 1.0, 1.2, 2.0, 2.4, 3.5])
@pytest.mark.parametrize("method", list(DiscreteSamplingMethod))
def test_discrete_sampling(fraction, method):
    """Size consistency tests for discrete sampling utility."""
    df = pd.DataFrame(np.random.randint(0, 100, size=(100, 4)), columns=list("ABCD"))

    n_points = math.ceil(fraction * len(df))
    sampled = sample_numerical_df(df, n_points, method=method)

    assert len(sampled) == n_points, (
        "Sampling did not return expected number of points."
    )
    if fraction >= 1.0:
        # Ensure the entire dataframe is contained in the sampled points
        assert (
            pd.merge(df, sampled, how="left", indicator=True)["_merge"].eq("both").all()
        ), "Oversized sampling did not return all original points at least once."
    else:
        # Assure all points are unique
        assert len(sampled) == len(sampled.drop_duplicates()), (
            "Undersized sampling did not return unique points."
        )


@given(
    points=hnp.arrays(
        dtype=float,
        shape=hnp.array_shapes(min_dims=2, max_dims=2, min_side=1),
        # Because of the square involved in the Euclidean distance computation,
        # we limit the floating point range to avoid overflow problems
        elements=st.floats(min_value=-1e100, max_value=1e100, allow_nan=False),
    )
)
# Explicitly test scenario with equidistant points (see comments in test body)
@pytest.mark.parametrize("random_tie_break", [False, True])
@example(points=np.array([[0, 0], [0, 1], [1, 0], [1, 1]]))
def test_farthest_point_sampling(points: np.ndarray, random_tie_break: bool):
    """FPS produces the same point sequence regardless of the order in which the
    points are provided. Also, each point fulfills the "farthest point" criterion
    in its respective iteration.
    """  # noqa
    # Order the points using FPS
    sorting_idxs = farthest_point_sampling(
        points, len(points), random_tie_break=random_tie_break
    )
    target = points[sorting_idxs]

    # For the ordered collection of points, it must hold:
    # ---------------------------------------------------
    # The minimum distance of the n_th selected point to all previously selected points
    # must be larger than the minimum distance of any other remaining candidate point to
    # the previously selected points – that's what makes it the "farthest point" in the
    # respective iteration.
    #
    # For the check, we start with the second point (because there are otherwise no
    # previous points) and end with the second last point (because there are otherwise
    # no alternative candidates left):
    dist_mat = pairwise_distances(target)
    for i in range(1, len(dist_mat) - 1):
        min_dist_selected_to_previous = np.min(dist_mat[i, :i])
        min_dist_remaining_to_previous = np.min(dist_mat[i + 1 :, :i])
        z = min_dist_selected_to_previous >= min_dist_remaining_to_previous
        assert z

    # Also, if the algorithm is set to fully deterministic, the obtained result should
    # not depend on the particular (random) order in which the points are provided.
    # That is, running the algorithm on a permutation should still produce the same
    # sequence of points. The flag `random_tie_break` can adjust the deterministic
    # behaviour. Note: We establish the check on the point coordinates and not the
    # selection index, because the latter can still differ in case of duplicated points.
    #
    # Examples where this can make a difference is three points forming an equilateral
    # triangle or four points spanning a unit cube. Here, tie-breaking operations such
    # as `np.max` can lead to different results depending on the order.
    permutation_idxs = np.random.permutation(len(points))
    sorting_idxs = farthest_point_sampling(
        points[permutation_idxs], len(points), random_tie_break=random_tie_break
    )
    if not random_tie_break:
        assert np.array_equal(target, points[permutation_idxs][sorting_idxs])

    # Because requesting a single point needs special treatment in FPS,
    # we test this as additional case
    sorting_idxs = farthest_point_sampling(points[permutation_idxs], 1)
    assert np.array_equal(target[[0]], points[permutation_idxs][sorting_idxs])


def test_farthest_point_sampling_pathological_case():
    """FPS executed on a degenerate point cloud raises a warning."""
    points = np.ones((3, 3))
    with pytest.warns(UserWarning, match="identical"):
        selection = farthest_point_sampling(points, 2)
    assert selection == [0, 1]



"""Bayesian recommenders."""

from baybe.recommenders.pure.bayesian.botorch import BotorchRecommender
from baybe.recommenders.pure.bayesian.sequential_greedy import (
    SequentialGreedyRecommender,
)

__all__ = [
    "BotorchRecommender",
    "SequentialGreedyRecommender",
]



"""Utilities for doc testing."""

import re
from pathlib import Path
from textwrap import dedent


def extract_code_blocks(path: str | Path, include_tilde: bool = True) -> list[str]:
    """Extract all python code blocks from the specified file."""
    contents = Path(path).read_text()
    pattern = (
        r"\s*(?:```|~~~)python\n*(.*?)\n*\s*(?:```|~~~)"
        if include_tilde
        else r"\s*```python\n*(.*?)\n*\s*```"
    )
    code_blocks = [dedent(c) for c in re.findall(pattern, contents, flags=re.DOTALL)]

    return code_blocks



"""Basic result classes for benchmarking."""

from __future__ import annotations

import sys

import importlib_metadata
from attrs import define, field
from attrs.validators import deep_mapping, instance_of
from cattrs.gen import make_dict_unstructure_fn
from pandas import DataFrame

from benchmarks.result import ResultMetadata
from benchmarks.serialization import BenchmarkSerialization, converter


@define(frozen=True)
class Result(BenchmarkSerialization):
    """A single benchmarking result."""

    benchmark_identifier: str = field(validator=instance_of(str))
    """The identifier of the benchmark that produced the result."""

    data: DataFrame = field(validator=instance_of(DataFrame))
    """The result of the benchmarked callable."""

    metadata: ResultMetadata = field(validator=instance_of(ResultMetadata))
    """The metadata associated with the benchmark result."""

    python_env: dict[str, str] = field(
        init=False,
        validator=deep_mapping(instance_of(str), instance_of(str), instance_of(dict)),
    )
    """The Python environment in which the benchmark was executed."""

    python_version: str = field(
        init=False, default=sys.version, validator=instance_of(str)
    )
    """The Python version with which the benchmark was executed."""

    @python_env.default
    def _default_python_env(self) -> dict[str, str]:
        installed_packages = importlib_metadata.distributions()
        return {dist.name: dist.version for dist in installed_packages}


converter.register_unstructure_hook(
    Result,
    make_dict_unstructure_fn(Result, converter, _cattrs_include_init_false=True),
)
